from draive.helpers.instruction_preparation import (
    InstructionPreparationAmbiguity,
    prepare_instruction,
)
from draive.helpers.instruction_refinement import refine_instruction
from draive.helpers.volatile_vector_index import VolatileVectorIndex

__all__ = (
    "InstructionPreparationAmbiguity",
    "VolatileVectorIndex",
    "prepare_instruction",
    "refine_instruction",
)



from collections.abc import Sequence
from typing import Any, Final, Protocol, Self, final, runtime_checkable

from haiway import State

__all__ = (
    "MEMORY_NONE",
    "Memory",
    "MemoryRecalling",
    "MemoryRemembering",
)


@runtime_checkable
class MemoryRecalling[Recalled](Protocol):
    async def __call__(
        self,
        **extra: Any,
    ) -> Recalled: ...


@runtime_checkable
class MemoryRemembering[Remembered](Protocol):
    async def __call__(
        self,
        *items: Remembered,
        **extra: Any,
    ) -> None: ...


async def _recall_none(
    **extra: Any,
) -> Any:
    return ()


async def _remember_none(
    *items: Any,
    **extra: Any,
) -> None:
    pass  # noop


@final
class Memory[Recalled, Remembered](State):
    @classmethod
    def constant(
        cls,
        recalled: Recalled,
    ) -> Self:
        async def recall(
            **extra: Any,
        ) -> Recalled:
            return recalled

        return cls(
            recall=recall,
            remember=_remember_none,
        )

    @classmethod
    def volatile[Item](
        cls,
        *,
        initial: Item,
    ) -> "Memory[Item, Item]":
        storage: Item = initial

        async def recall(
            **extra: Any,
        ) -> Item:
            nonlocal storage
            return storage

        async def remember(
            *items: Item,
            **extra: Any,
        ) -> None:
            nonlocal storage
            if not items:
                return  # nothing to do

            storage = items[-1]

        return Memory[Item, Item](
            recall=recall,
            remember=remember,
        )

    @classmethod
    def accumulative_volatile[Item](
        cls,
        *,
        initial: Sequence[Item] | None = None,
        limit: int | None = None,
    ) -> "Memory[Sequence[Item], Item]":
        storage: Sequence[Item] = tuple(initial) if initial else ()
        limit = limit or 0

        async def recall(
            **extra: Any,
        ) -> Sequence[Item]:
            nonlocal storage
            return storage

        async def remember(
            *items: Item,
            **extra: Any,
        ) -> None:
            nonlocal storage
            if not items:
                return  # nothing to do

            storage = (*storage, *items)

            if limit > 0:
                storage = tuple(storage[-limit:])

        return Memory[Sequence[Item], Item](
            recall=recall,
            remember=remember,
        )

    recall: MemoryRecalling[Recalled]
    remember: MemoryRemembering[Remembered]


MEMORY_NONE: Final[Memory[Any, Any]] = Memory[Any, Any](
    recall=_recall_none,
    remember=_remember_none,
)



from collections.abc import Generator, Sequence
from uuid import uuid4

from google.genai.types import (
    ContentDict,
    Part,
    PartDict,
)
from haiway import META_EMPTY, as_dict

from draive.lmm import (
    LMMCompletion,
    LMMContextElement,
    LMMInput,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponses,
)
from draive.multimodal import (
    MediaData,
    MediaReference,
    MetaContent,
    MultimodalContentElement,
    TextContent,
)

__all__ = (
    "context_element_as_content",
    "result_part_as_content_or_call",
)


def context_element_as_content(
    element: LMMContextElement,
    /,
) -> ContentDict:
    if isinstance(element, LMMInput):
        return {
            "role": "user",
            "parts": list(content_parts(element.content.parts)),
        }

    elif isinstance(element, LMMCompletion):
        return {
            "role": "model",
            "parts": list(content_parts(element.content.parts)),
        }

    elif isinstance(element, LMMToolRequests):
        return {
            "role": "model",
            "parts": [
                {
                    "function_call": {
                        "id": request.identifier,
                        "name": request.tool,
                        "args": as_dict(request.arguments),
                    }
                }
                for request in element.requests
            ],
        }

    elif isinstance(element, LMMToolResponses):
        return {
            "role": "user",
            "parts": [
                {
                    "function_response": {
                        "id": response.identifier,
                        "name": response.tool,
                        "response": {
                            "error": list(content_parts(response.content.parts)),
                        }
                        if response.handling == "error"
                        else {
                            "output": list(content_parts(response.content.parts)),
                        },
                    }
                }
                for response in element.responses
            ],
        }


def content_parts(
    elements: Sequence[MultimodalContentElement],
    /,
) -> Generator[PartDict]:
    for element in elements:
        if isinstance(element, TextContent):
            yield {
                "text": element.text,
            }

        elif isinstance(element, MediaData):
            yield {
                "inline_data": {
                    "data": element.data,
                    "mime_type": element.media,
                },
            }

        elif isinstance(element, MediaReference):
            yield {
                "file_data": {
                    "file_uri": element.uri,
                    "mime_type": element.media,
                }
            }

        elif isinstance(element, MetaContent):
            if element.category == "thought" and element.content:
                yield {
                    "text": element.content.to_str(),
                    "thought": True,
                }

            elif element.category == "transcript" and element.content:
                yield {
                    "text": element.content.to_str(),
                }

            else:
                continue  # skip other meta

        else:
            yield {
                "text": element.to_json(),
            }


def result_part_as_content_or_call(
    part: Part,
    /,
) -> Generator[MultimodalContentElement | LMMToolRequest]:
    if part.text:
        # assuming only text thinking is possible
        if part.thought:
            yield MetaContent.of(
                "thought",
                content=TextContent.of(part.text),
                meta=META_EMPTY,
            )

        else:
            yield TextContent.of(part.text)

    if part.function_call and part.function_call.name:  # can't call without a name
        yield LMMToolRequest(
            identifier=part.function_call.id or uuid4().hex,
            tool=part.function_call.name,
            arguments=part.function_call.args or {},
        )

    if part.inline_data and part.inline_data.data:  # there is no content without content...
        yield MediaData.of(
            part.inline_data.data,
            media=part.inline_data.mime_type or "application/octet-stream",
        )

    if part.file_data and part.file_data.file_uri:  # there is no content without content...
        yield MediaReference.of(
            part.file_data.file_uri,
            media=part.file_data.mime_type or "application/octet-stream",
        )



from asyncio import Task, wait
from asyncio.tasks import FIRST_COMPLETED
from collections.abc import AsyncIterator, MutableSet, Sequence
from datetime import UTC, datetime
from typing import Any, Literal, overload
from uuid import UUID, uuid4

from haiway import AsyncQueue, ctx

from draive.conversation.completion.types import ConversationMemory, ConversationMessage
from draive.conversation.types import (
    ConversationElement,
    ConversationEvent,
    ConversationMessageChunk,
    ConversationStreamElement,
)
from draive.guardrails import GuardrailsModeration
from draive.instructions import Instruction
from draive.lmm import (
    LMM,
    LMMCompletion,
    LMMContextElement,
    LMMInstruction,
    LMMStreamChunk,
    LMMToolRequest,
    LMMToolRequests,
    LMMToolResponse,
    LMMToolResponses,
)
from draive.multimodal import MultimodalContent
from draive.tools import Toolbox
from draive.utils import Processing, ProcessingEvent

__all__ = ("conversation_completion",)


@overload
async def conversation_completion(
    *,
    instruction: Instruction | None,
    input: ConversationMessage,
    memory: ConversationMemory,
    toolbox: Toolbox,
    stream: Literal[False] = False,
    **extra: Any,
) -> ConversationMessage: ...


@overload
async def conversation_completion(
    *,
    instruction: Instruction | None,
    input: ConversationMessage,
    memory: ConversationMemory,
    toolbox: Toolbox,
    stream: Literal[True],
    **extra: Any,
) -> AsyncIterator[ConversationStreamElement]: ...


async def conversation_completion(
    *,
    instruction: Instruction | None,
    input: ConversationMessage,  # noqa: A002
    memory: ConversationMemory,
    toolbox: Toolbox,
    stream: bool = False,
    **extra: Any,
) -> AsyncIterator[ConversationStreamElement] | ConversationMessage:
    async with ctx.scope("conversation_completion"):
        # relying on memory recall correctness
        recalled_messages: Sequence[ConversationElement] = await memory.recall()
        await GuardrailsModeration.check_input(input.content)
        context: list[LMMContextElement] = [
            *(
                message.to_lmm_context()
                for message in recalled_messages
                if isinstance(message, ConversationMessage)
            ),
            input.to_lmm_context(),
        ]

        if stream:
            return await _conversation_completion_stream(
                instruction=instruction,
                input=input,
                context=context,
                memory=memory,
                toolbox=toolbox,
            )

        else:
            return await _conversation_completion(
                instruction=instruction,
                input=input,
                context=context,
                memory=memory,
                toolbox=toolbox,
            )


async def _conversation_completion(
    instruction: Instruction | None,
    input: ConversationMessage,  # noqa: A002
    context: list[LMMContextElement],
    memory: ConversationMemory,
    toolbox: Toolbox,
    **extra: Any,
) -> ConversationMessage:
    formatted_instruction: LMMInstruction | None = Instruction.formatted(instruction)
    tools_turn: int = 0
    result: LMMCompletion
    result_extension: MultimodalContent = MultimodalContent.empty
    while True:
        ctx.log_debug("...requesting completion...")
        match await LMM.completion(
            instruction=formatted_instruction,
            context=context,
            tools=toolbox.available_tools(tools_turn=tools_turn),
            **extra,
        ):
            case LMMCompletion() as completion:
                ctx.log_debug("...received result...")
                if result_extension:
                    result = completion.updated(
                        content=result_extension.appending(completion.content)
                    )

                else:
                    result = completion

                break  # proceed to resolving

            case LMMToolRequests() as tool_requests:
                ctx.log_debug(f"...received tool requests (turn {tools_turn})...")
                # skip tool_requests.content - no need for extra comments
                tool_responses: LMMToolResponses = await toolbox.respond_all(tool_requests)

                if completion := tool_responses.completion(extension=result_extension):
                    ctx.log_debug("...received tools direct result...")
                    result = completion
                    break  # proceed to resolving

                elif extension := tool_responses.completion_extension():
                    ctx.log_debug("...received tools result extension...")
                    result_extension = result_extension.appending(extension)

                ctx.log_debug("...received tools responses...")
                context.extend((tool_requests, tool_responses))

        tools_turn += 1  # continue with next turn

    ctx.log_debug("...finalizing message...")
    response_message: ConversationMessage = ConversationMessage.model(
        created=datetime.now(UTC),
        content=result,
        meta=result.meta,
    )

    ctx.log_debug("...remembering...")
    await memory.remember(input, response_message)

    ctx.log_debug("... response message finished!")
    return response_message


async def _conversation_completion_stream(  # noqa: C901, PLR0915
    instruction: Instruction | None,
    input: ConversationMessage,  # noqa: A002
    context: list[LMMContextElement],
    memory: ConversationMemory,
    toolbox: Toolbox,
    **extra: Any,
) -> AsyncIterator[ConversationStreamElement]:
    output_queue = AsyncQueue[ConversationStreamElement]()
    # completion message identifier to match chunks and events
    message_identifier: UUID = uuid4()

    async def report_event(
        event: ProcessingEvent,
        /,
    ) -> None:
        output_queue.enqueue(
            ConversationEvent.of(
                event.name,
                identifier=event.identifier,
                content=event.content,
                meta=event.meta,
            )
        )

    with ctx.updated(ctx.state(Processing).updated(event_reporting=report_event)):

        async def consume_lmm_output() -> None:  # noqa: C901, PLR0912, PLR0915
            tools_turn: int = 0
            try:
                formatted_instruction: LMMInstruction | None = Instruction.formatted(instruction)
                result: MultimodalContent = MultimodalContent.empty
                while True:
                    pending_tool_requests: list[LMMToolRequest] = []
                    pending_tool_responses: MutableSet[Task[LMMToolResponse]] = set()
                    accumulated_content: MultimodalContent = MultimodalContent.empty
                    ctx.log_debug("...requesting completion...")
                    async for element in await LMM.completion(
                        instruction=formatted_instruction,
                        context=context,
                        tools=toolbox.available_tools(tools_turn=tools_turn),
                        output="text",
                        stream=True,
                        **extra,
                    ):
                        match element:
                            case LMMStreamChunk() as chunk:
                                ctx.log_debug("...received result chunk...")
                                # we are ending the stream when all tool results are provided
                                # and a chunk is marked as final, final mark before all tool
                                # calls are resolved typically means that there will be no more
                                # tool related content, also models may produce content and tools
                                if chunk.eod and not pending_tool_requests:
                                    result = result.appending(
                                        accumulated_content.appending(chunk.content)
                                    )
                                    output_queue.enqueue(
                                        ConversationMessageChunk.model(
                                            message_identifier=message_identifier,
                                            content=chunk.content,
                                            eod=True,
                                        )
                                    )
                                    output_queue.finish()
                                    break  # proceed to finalization (outer loop)

                                elif chunk.content:  # skip empty chunks
                                    accumulated_content = accumulated_content.appending(
                                        chunk.content
                                    )
                                    output_queue.enqueue(
                                        ConversationMessageChunk.model(
                                            message_identifier=message_identifier,
                                            content=chunk.content,
                                            eod=False,
                                        )
                                    )

                            case LMMToolRequest() as tool_request:
                                ctx.log_debug(f"...received tool requests (turn {tools_turn})...")
                                pending_tool_requests.append(tool_request)
                                # start processing immediately
                                pending_tool_responses.add(ctx.spawn(toolbox.respond, tool_request))

                    if not pending_tool_responses or not pending_tool_requests:
                        break  # proceed to finalization

                    tool_requests: LMMToolRequests = LMMToolRequests.of(
                        pending_tool_requests,
                        content=accumulated_content,
                    )

                    tools_completion: bool = False
                    responses: list[LMMToolResponse] = []
                    while pending_tool_responses:
                        completed, pending_tool_responses = await wait(
                            pending_tool_responses,
                            return_when=FIRST_COMPLETED,
                        )
                        for response in completed:
                            tool_response: LMMToolResponse = response.result()
                            if tool_response.handling == "completion":
                                ctx.log_debug("...received tools direct result...")
                                output_queue.enqueue(
                                    ConversationMessageChunk.model(
                                        message_identifier=message_identifier,
                                        content=tool_response.content,
                                        eod=False,
                                    )
                                )
                                tools_completion = True
                                result = result.appending(tool_response.content)

                            elif tool_response.handling == "extension":
                                ctx.log_debug("...received tools result extension...")
                                output_queue.enqueue(
                                    ConversationMessageChunk.model(
                                        message_identifier=message_identifier,
                                        content=tool_response.content,
                                        eod=False,
                                    )
                                )
                                responses.append(tool_response)
                                result = result.appending(tool_response.content)

                            else:
                                responses.append(tool_response)

                    if tools_completion:
                        output_queue.enqueue(  # send empty end of data part
                            ConversationMessageChunk.model(
                                message_identifier=message_identifier,
                                content=MultimodalContent.empty,
                                eod=True,
                            )
                        )
                        output_queue.finish()
                        break  # proceed to finalization

                    ctx.log_debug("...received tools responses...")
                    tool_responses: LMMToolResponses = LMMToolResponses.of(responses)
                    context.extend((tool_requests, tool_responses))

                    tools_turn += 1  # continue with next turn

            except BaseException as exc:
                output_queue.finish(exception=exc)
                raise exc

            assert output_queue.is_finished  # nosec: B101

            ctx.log_debug("...finalizing message...")
            response_message: ConversationMessage = ConversationMessage.model(
                created=datetime.now(UTC),
                identifier=message_identifier,
                content=result,
            )

            ctx.log_debug("...remembering...")
            await memory.remember(input, response_message)

            ctx.log_debug("... response message finished!")

        ctx.spawn(consume_lmm_output)

        return output_queue



import json
from collections.abc import Mapping, Sequence
from pathlib import Path
from string import Formatter
from typing import Any, final

from haiway import Meta, asynchronous, ctx

from draive.instructions.types import (
    Instruction,
    InstructionDeclaration,
    InstructionDeclarationArgument,
)

__all__ = ("InstructionsFileStorage",)


@final
class InstructionsFileStorage:
    def __init__(
        self,
        path: Path | str,
    ) -> None:
        self._path: Path
        match path:
            case str() as path_str:
                self._path = Path(path_str)

            case path:
                self._path = path

        self._storage: Mapping[str, str] | None = None
        self._listing: Sequence[InstructionDeclaration] | None = None

    async def fetch_list(
        self,
        **extra: Any,
    ) -> Sequence[InstructionDeclaration]:
        if self._listing is not None:
            return self._listing

        if self._storage is None:
            self._storage = await self._file_load()

        formatter = Formatter()
        self._listing = tuple(
            InstructionDeclaration(
                name=name,
                arguments=tuple(
                    InstructionDeclarationArgument(name=arg_name)
                    for _, arg_name, _, _ in formatter.parse(content)
                    if arg_name  # we could also check for positional arguments
                ),
                meta=Meta({"file": str(self._path)}),
            )
            for name, content in self._storage.items()
        )

        return self._listing

    async def fetch_instruction(
        self,
        name: str,
        *,
        arguments: Mapping[str, str | float | int] | None = None,
        **extra: Any,
    ) -> Instruction | None:
        if self._storage is None:
            self._storage = await self._file_load()

        if instruction := self._storage.get(name):
            return Instruction.of(
                instruction,
                name=name,
                meta={"file": str(self._path)},
                arguments=arguments,
            )

        else:
            ctx.log_debug(f"{self._path} does not contain instruction '{name}'")
            return None

    @asynchronous
    def _file_load(
        self,
    ) -> Mapping[str, str]:
        if self._path.exists():
            with open(self._path, mode="rb") as file:
                match json.loads(file.read()):
                    case {**elements}:
                        return {
                            key: value for key, value in elements.items() if isinstance(value, str)
                        }

                    case _:
                        return {}

        else:
            return {}



from collections.abc import AsyncIterable, Iterable
from typing import Any, Protocol, runtime_checkable

from draive.instructions import Instruction
from draive.multimodal import MultimodalContent
from draive.prompts import Prompt
from draive.tools import Toolbox

__all__ = ("TextGenerating",)


@runtime_checkable
class TextGenerating(Protocol):
    async def __call__(
        self,
        *,
        instruction: Instruction | None,
        input: Prompt | MultimodalContent,  # noqa: A002
        toolbox: Toolbox,
        examples: Iterable[tuple[MultimodalContent, str]],
        **extra: Any,
    ) -> AsyncIterable[str] | str: ...



from collections.abc import Iterable
from typing import Any, final

from haiway import State, as_tuple, ctx

from draive.conversation.realtime.default import realtime_conversation_preparing
from draive.conversation.realtime.types import (
    RealtimeConversationPreparing,
    RealtimeConversationSessionScope,
)
from draive.conversation.types import ConversationMemory, ConversationMessage
from draive.instructions import Instruction
from draive.tools import Tool, Toolbox
from draive.utils import Memory

__all__ = ("RealtimeConversation",)


@final
class RealtimeConversation(State):
    @classmethod
    async def prepare(
        cls,
        *,
        instruction: Instruction | None = None,
        memory: ConversationMemory | Iterable[ConversationMessage] | None = None,
        tools: Toolbox | Iterable[Tool] | None = None,
        **extra: Any,
    ) -> RealtimeConversationSessionScope:
        conversation: RealtimeConversation = ctx.state(cls)

        # prepare memory
        conversation_memory: ConversationMemory | None
        match memory:
            case None:
                conversation_memory = None

            case Memory() as memory:
                conversation_memory = memory

            case memory_messages:
                conversation_memory = ConversationMemory.constant(as_tuple(memory_messages))

        return await conversation.preparing(
            instruction=Instruction.of(instruction),
            memory=conversation_memory,
            toolbox=Toolbox.of(tools),
            **extra,
        )

    preparing: RealtimeConversationPreparing = realtime_conversation_preparing



from collections.abc import AsyncIterable, Iterable
from typing import Any

from haiway import ctx

from draive.instructions import Instruction
from draive.lmm import (
    LMM,
    LMMCompletion,
    LMMContextElement,
    LMMInput,
    LMMInstruction,
    LMMToolRequests,
    LMMToolResponses,
)
from draive.multimodal import MultimodalContent
from draive.prompts import Prompt
from draive.tools import Toolbox

__all__ = ("generate_text",)


async def generate_text(
    *,
    instruction: Instruction | None,
    input: Prompt | MultimodalContent,  # noqa: A002
    toolbox: Toolbox,
    examples: Iterable[tuple[MultimodalContent, str]],
    **extra: Any,
) -> AsyncIterable[str] | str:
    async with ctx.scope("generate_text"):
        context: list[LMMContextElement] = [
            *[
                message
                for example in examples
                for message in [
                    LMMInput.of(example[0]),
                    LMMCompletion.of(example[1]),
                ]
            ],
        ]

        match input:
            case Prompt() as prompt:
                context.extend(prompt.content)

            case value:
                context.append(LMMInput.of(value))

        return await _text_generation(
            instruction=instruction,
            context=context,
            toolbox=toolbox,
        )


async def _text_generation(
    instruction: Instruction | None,
    context: list[LMMContextElement],
    toolbox: Toolbox,
    **extra: Any,
) -> str:
    formatted_instruction: LMMInstruction | None = Instruction.formatted(instruction)
    tools_turn: int = 0
    result: MultimodalContent = MultimodalContent.empty
    result_extension: MultimodalContent = MultimodalContent.empty
    while True:
        ctx.log_debug("...requesting completion...")
        match await LMM.completion(
            instruction=formatted_instruction,
            context=context,
            tools=toolbox.available_tools(tools_turn=tools_turn),
            output="text",
            **extra,
        ):
            case LMMCompletion() as completion:
                ctx.log_debug("...received result...")
                result = result_extension.appending(completion.content)
                break  # proceed to resolving

            case LMMToolRequests() as tool_requests:
                ctx.log_debug(f"...received tool requests (turn {tools_turn})...")
                # skip tool_requests.content - no need for extra comments
                tool_responses: LMMToolResponses = await toolbox.respond_all(tool_requests)

                if completion := tool_responses.completion(extension=result_extension):
                    ctx.log_debug("...received tools direct result...")
                    result = completion.content
                    break  # proceed to resolving

                elif extension := tool_responses.completion_extension():
                    ctx.log_debug("...received tools result extension...")
                    result_extension = result_extension.appending(extension)

                ctx.log_debug("...received tools responses...")
                context.extend((tool_requests, tool_responses))

        tools_turn += 1  # continue with next turn

    ctx.log_debug("...text completed!")
    return result.to_str()



import random
from asyncio import Lock
from collections.abc import Callable, Collection, Iterable, Sequence
from pathlib import Path
from typing import Protocol, Self, runtime_checkable
from uuid import UUID, uuid4

from haiway import (
    META_EMPTY,
    Meta,
    MetaValues,
    State,
    as_list,
    asynchronous,
    ctx,
    execute_concurrently,
)

from draive.evaluation.evaluator import EvaluatorResult, PreparedEvaluator
from draive.evaluation.generator import generate_case_parameters
from draive.evaluation.scenario import EvaluatorScenarioResult, PreparedEvaluatorScenario
from draive.parameters import DataModel, Field

__all__ = (
    "EvaluatorCaseResult",
    "EvaluatorSuite",
    "EvaluatorSuiteCaseResult",
    "EvaluatorSuiteDefinition",
    "EvaluatorSuiteResult",
    "EvaluatorSuiteStorage",
    "evaluator_suite",
)


class EvaluatorSuiteCase[CaseParameters: DataModel](DataModel):
    """
    Individual test case for an evaluator suite.

    Represents a single test case with parameters, unique identifier,
    and optional comment for documentation.

    Attributes
    ----------
    identifier : str
        Unique identifier for the test case
    parameters : CaseParameters
        Parameters specific to this test case
    comment : str | None
        Optional comment describing the test case
    """

    identifier: str = Field(default_factory=lambda: str(uuid4()))
    parameters: CaseParameters
    comment: str | None = None


class EvaluatorSuiteCaseResult[CaseParameters: DataModel](DataModel):
    """
    Result of evaluating a single test case.

    Contains the test case and its evaluation results, with methods to
    check if the case passed and generate reports.

    Attributes
    ----------
    case : EvaluatorSuiteCase[CaseParameters]
        The test case that was evaluated
    results : Sequence[EvaluatorScenarioResult]
        Results from all scenario evaluations
    meta : Meta
        Additional metadata for the case result
    """

    case: EvaluatorSuiteCase[CaseParameters] = Field(
        description="Evaluated case",
    )
    results: Sequence[EvaluatorScenarioResult] = Field(
        description="Evaluation results",
    )
    meta: Meta = Field(
        description="Additional evaluation metadata",
        default=META_EMPTY,
    )

    @property
    def passed(self) -> bool:
        """
        Check if all scenario evaluations passed.

        Returns
        -------
        bool
            True if all scenario evaluations passed, False otherwise.
            Empty results return False.
        """
        # empty results is equivalent of failure
        return len(self.results) > 0 and all(result.passed for result in self.results)

    def report(
        self,
        *,
        detailed: bool = True,
        include_passed: bool = True,
    ) -> str:
        """
        Generate a human-readable report of the case results.

        Parameters
        ----------
        detailed : bool, optional
            If True, include full details in XML format. If False, return
            brief summary, by default True
        include_passed : bool, optional
            If True, include passed evaluations in the report. If False,
            only show failed evaluations, by default True

        Returns
        -------
        str
            Formatted report string
        """
        if not self.results:
            return f"{self.case.identifier} has no evaluations"

        status: str = "passed" if self.passed else "failed"
        results_report: str = "\n".join(
            result.report(
                detailed=detailed,
                include_passed=include_passed,
            )
            for result in self.results
            if include_passed or not result.passed
        )

        if not detailed:
            if results_report:
                return f"{self.case.identifier}: {status} ({self.performance}%)\n{results_report}"

            else:
                return f"{self.case.identifier}: {status} ({self.performance}%)"

        if results_report:
            return (
                f"<evaluator_suite case='{self.case.identifier}' status='{status}'"
                f" performance='{self.performance:.2f}%'>"
                f"\n<results>\n{results_report}\n</results>"
                "\n</evaluator_suite>"
            )

        else:
            return f"<evaluator_suite case='{self.case.identifier}' status='{status}' />"

    @property
    def performance(self) -> float:
        """
        Calculate average performance across all scenario evaluations.

        Returns
        -------
        float
            Average performance percentage (0-100) across all evaluations.
            Returns 0.0 if no evaluations.
        """
        if not self.results:
            return 0.0

        score: float = 0.0
        for evaluation in self.results:
            score += evaluation.performance

        return score / len(self.results)


class EvaluatorSuiteResult[SuiteParameters: DataModel, CaseParameters: DataModel](DataModel):
    parameters: SuiteParameters
    cases: Sequence[EvaluatorSuiteCaseResult[CaseParameters]]

    @property
    def passed(self) -> bool:
        return all(case.passed for case in self.cases)

    def report(
        self,
        *,
        detailed: bool = True,
        include_passed: bool = True,
    ) -> str:
        if not self.cases:
            return "Evaluator suite empty!"

        status: str = "passed" if self.passed else "failed"
        results_report: str = "\n---\n".join(
            result.report(
                detailed=detailed,
                include_passed=include_passed,
            )
            for result in self.cases
            if include_passed or not result.passed
        )

        if not detailed:
            if results_report:
                return f"Evaluator suite: {status} ({self.performance:.2f}%):\n\n{results_report}"

            else:
                return f"Evaluator suite: {status} ({self.performance:.2f}%)"

        if results_report:
            return (
                f"<evaluator_suite status='{status}'"
                f" performance='{self.performance:.2f}%'>"
                f"\n<results>\n{results_report}\n</results>"
                "\n</evaluator_suite>"
            )

        else:
            return f"<evaluator_suite status='{status}' performance='{self.performance:.2f}%' />"

    @property
    def performance(self) -> float:
        if not self.cases:
            return 0.0

        score: float = 0.0
        for evaluation in self.cases:
            score += evaluation.performance

        return score / len(self.cases)


class EvaluatorCaseResult(DataModel):
    @classmethod
    def of(
        cls,
        result: EvaluatorScenarioResult | EvaluatorResult,
        *results: EvaluatorScenarioResult | EvaluatorResult,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        free_results: list[EvaluatorResult] = []
        scenario_results: list[EvaluatorScenarioResult] = []
        for element in (result, *results):
            if isinstance(element, EvaluatorScenarioResult):
                scenario_results.append(element)

            else:
                free_results.append(element)

        if free_results:
            scenario_results.append(
                EvaluatorScenarioResult(
                    scenario="EvaluatorSuite",
                    evaluations=free_results,
                )
            )

        return cls(
            results=scenario_results,
            meta=Meta.of(meta),
        )

    @classmethod
    async def evaluating[Value](
        cls,
        value: Value,
        /,
        evaluator: PreparedEvaluatorScenario[Value] | PreparedEvaluator[Value],
        *evaluators: PreparedEvaluatorScenario[Value] | PreparedEvaluator[Value],
        concurrent_tasks: int = 2,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        async def execute(
            evaluator: PreparedEvaluatorScenario[Value] | PreparedEvaluator[Value],
        ) -> EvaluatorScenarioResult | EvaluatorResult:
            return await evaluator(value)

        return cls.of(
            *await execute_concurrently(
                execute,
                (evaluator, *evaluators),
                concurrent_tasks=concurrent_tasks,
            ),
            meta=Meta.of(meta),
        )

    @classmethod
    def merging(
        cls,
        result: Self,
        *results: Self,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        merged_evaluations: list[EvaluatorScenarioResult] = as_list(result.results)
        merged_meta: Meta = result.meta
        for other in results:
            merged_evaluations.extend(other.results)
            merged_meta = merged_meta.merged_with(other.meta)

        return cls(
            results=merged_evaluations,
            meta=merged_meta.merged_with(Meta.of(meta)),
        )

    results: Sequence[EvaluatorScenarioResult] = Field(
        description="Evaluation results",
    )
    meta: Meta = Field(
        description="Additional evaluation metadata",
        default=META_EMPTY,
    )


@runtime_checkable
class EvaluatorSuiteDefinition[
    SuiteParameters: DataModel,
    CaseParameters: DataModel,
](Protocol):
    async def __call__(
        self,
        parameters: SuiteParameters,
        case_parameters: CaseParameters,
    ) -> EvaluatorCaseResult: ...


class EvaluatorSuiteData[SuiteParameters: DataModel, CaseParameters: DataModel](DataModel):
    parameters: SuiteParameters
    cases: Sequence[EvaluatorSuiteCase[CaseParameters]] = ()


@runtime_checkable
class EvaluatorSuiteStorage[SuiteParameters: DataModel, CaseParameters: DataModel](Protocol):
    async def load(
        self,
    ) -> EvaluatorSuiteData[SuiteParameters, CaseParameters]: ...

    async def save(
        self,
        data: EvaluatorSuiteData[SuiteParameters, CaseParameters],
    ) -> None: ...


class EvaluatorSuite[
    SuiteParameters: DataModel,
    CaseParameters: DataModel,
]:
    __slots__ = (
        "_case_parameters",
        "_data_cache",
        "_definition",
        "_lock",
        "_state",
        "_storage",
        "_suite_parameters",
    )

    def __init__(
        self,
        suite_parameters: type[SuiteParameters],
        case_parameters: type[CaseParameters],
        definition: EvaluatorSuiteDefinition[SuiteParameters, CaseParameters],
        storage: EvaluatorSuiteStorage[SuiteParameters, CaseParameters],
        state: Collection[State],
    ) -> None:
        self._suite_parameters: type[SuiteParameters] = suite_parameters
        self._case_parameters: type[CaseParameters] = case_parameters
        self._definition: EvaluatorSuiteDefinition[SuiteParameters, CaseParameters] = definition
        self._storage: EvaluatorSuiteStorage[SuiteParameters, CaseParameters] = storage
        self._state: Collection[State] = state
        self._data_cache: EvaluatorSuiteData[SuiteParameters, CaseParameters] | None = None
        self._lock: Lock = Lock()

    async def __call__(
        self,
        case_parameters: Sequence[EvaluatorSuiteCase[CaseParameters] | CaseParameters | UUID]
        | float
        | int
        | None = None,
        /,
        parameters: SuiteParameters | None = None,
        concurrent_cases: int = 2,
        reload: bool = False,
    ) -> EvaluatorSuiteResult[SuiteParameters, CaseParameters]:
        async with ctx.scope("evaluation.suite", *self._state):
            return await self._evaluate(
                case_parameters,
                suite_parameters=parameters,
                concurrent_cases=concurrent_cases,
                reload=reload,
            )

    async def _evaluate(
        self,
        case_parameters: Sequence[EvaluatorSuiteCase[CaseParameters] | CaseParameters | UUID]
        | float
        | int
        | None,
        /,
        *,
        suite_parameters: SuiteParameters | None,
        concurrent_cases: int,
        reload: bool,
    ) -> EvaluatorSuiteResult[SuiteParameters, CaseParameters]:
        suite_data: EvaluatorSuiteData[SuiteParameters, CaseParameters]
        async with self._lock:
            suite_data = await self._data(reload=reload)

        cases: Sequence[EvaluatorSuiteCase[CaseParameters]] = []
        if not case_parameters:
            cases = suite_data.cases

        elif isinstance(case_parameters, int):
            cases = random.sample(  # nosec: B311
                suite_data.cases,
                min(len(suite_data.cases), case_parameters),
            )

        elif isinstance(case_parameters, float):
            assert 0 < case_parameters <= 1  # nosec: B101
            cases = random.sample(  # nosec: B311
                suite_data.cases,
                min(len(suite_data.cases), int(len(suite_data.cases) * case_parameters)),
            )

        else:
            cases = []
            for case in case_parameters:
                match case:
                    case UUID() as case_id:
                        if evaluation_case := next(
                            iter(case for case in suite_data.cases if case.identifier == case_id),
                            None,
                        ):
                            cases.append(evaluation_case)

                        else:
                            raise ValueError(f"Evaluation case with ID {case_id} does not exists.")

                    case EvaluatorSuiteCase():
                        cases.append(case)

                    case _:
                        cases.append(EvaluatorSuiteCase[CaseParameters](parameters=case))

        parameters: SuiteParameters = suite_parameters or suite_data.parameters

        async def evaluate_case(
            case: EvaluatorSuiteCase[CaseParameters],
        ) -> EvaluatorSuiteCaseResult[CaseParameters]:
            return await self._evaluate_case(
                case,
                suite_parameters=parameters,
            )

        return EvaluatorSuiteResult(
            parameters=parameters,
            cases=await execute_concurrently(
                evaluate_case,
                cases,
                concurrent_tasks=concurrent_cases,
            ),
        )

    async def _evaluate_case(
        self,
        case_parameters: EvaluatorSuiteCase[CaseParameters],
        *,
        suite_parameters: SuiteParameters,
    ) -> EvaluatorSuiteCaseResult[CaseParameters]:
        result: EvaluatorCaseResult = await self._definition(
            parameters=suite_parameters,
            case_parameters=case_parameters.parameters,
        )

        return EvaluatorSuiteCaseResult[CaseParameters](
            case=case_parameters,
            results=result.results,
        )

    async def _data(
        self,
        reload: bool = False,
    ) -> EvaluatorSuiteData[SuiteParameters, CaseParameters]:
        if reload or self._data_cache is None:
            self._data_cache = await self._storage.load()
            return self._data_cache

        else:
            return self._data_cache

    def with_state(
        self,
        state: State,
        /,
        *states: State,
    ) -> Self:
        return self.__class__(
            suite_parameters=self._suite_parameters,
            case_parameters=self._case_parameters,
            definition=self._definition,
            storage=self._storage,
            state=(*self._state, state, *states),
        )

    def with_storage(
        self,
        storage: EvaluatorSuiteStorage[SuiteParameters, CaseParameters]
        | EvaluatorSuiteData[SuiteParameters, CaseParameters]
        | Path
        | str,
        /,
    ) -> Self:
        suite_storage: EvaluatorSuiteStorage[SuiteParameters, CaseParameters]
        match storage:
            case str() as path_str:
                suite_storage = _EvaluatorSuiteFileStorage[SuiteParameters, CaseParameters](
                    path=path_str,
                    data_type=EvaluatorSuiteData[self._suite_parameters, self._case_parameters],
                )

            case Path() as path:
                suite_storage = _EvaluatorSuiteFileStorage[SuiteParameters, CaseParameters](
                    path=path,
                    data_type=EvaluatorSuiteData[self._suite_parameters, self._case_parameters],
                )

            case EvaluatorSuiteData() as data:
                suite_storage = _EvaluatorSuiteMemoryStorage[SuiteParameters, CaseParameters](
                    data_type=EvaluatorSuiteData[SuiteParameters, CaseParameters],
                    parameters=data.parameters,
                    cases=data.cases,
                )

            case storage:
                suite_storage = storage

        return self.__class__(
            suite_parameters=self._suite_parameters,
            case_parameters=self._case_parameters,
            definition=self._definition,
            storage=suite_storage,
            state=self._state,
        )

    async def cases(
        self,
        reload: bool = False,
    ) -> Sequence[EvaluatorSuiteCase[CaseParameters]]:
        async with self._lock:
            return (await self._data(reload=reload)).cases

    async def add_case(
        self,
        parameters: CaseParameters,
        /,
        comment: str | None = None,
    ) -> None:
        async with self._lock:
            data: EvaluatorSuiteData[SuiteParameters, CaseParameters] = await self._data(
                reload=True
            )
            self._data_cache = data.updated(
                cases=(
                    *data.cases,
                    EvaluatorSuiteCase[CaseParameters](
                        parameters=parameters,
                        comment=comment,
                    ),
                )
            )
            await self._storage.save(self._data_cache)

    async def generate_cases(
        self,
        *,
        count: int,
        persist: bool = False,
        guidelines: str | None = None,
        examples: Iterable[CaseParameters] | None = None,
    ) -> Sequence[EvaluatorSuiteCase[CaseParameters]]:
        async with self._lock:
            current: EvaluatorSuiteData[SuiteParameters, CaseParameters] = await self._data(
                reload=True
            )

            generated: Sequence[EvaluatorSuiteCase[CaseParameters]] = [
                EvaluatorSuiteCase(parameters=parameters)
                for parameters in await generate_case_parameters(
                    self._case_parameters,
                    count=count,
                    examples=examples or [case.parameters for case in current.cases],
                    guidelines=guidelines,
                )
            ]

            self._data_cache = current.updated(cases=(*current.cases, *generated))

            if persist:
                await self._storage.save(self._data_cache)

            return generated

    async def remove_case(
        self,
        identifier: UUID,
        /,
    ) -> None:
        async with self._lock:
            data: EvaluatorSuiteData[SuiteParameters, CaseParameters] = await self._data(
                reload=True
            )
            self._data_cache = data.updated(
                cases=tuple(case for case in data.cases if case.identifier != identifier)
            )
            await self._storage.save(self._data_cache)


def evaluator_suite[SuiteParameters: DataModel, CaseParameters: DataModel, Value](
    case_parameters: type[CaseParameters],
    /,
    suite_parameters: type[SuiteParameters] | SuiteParameters,
    *,
    storage: EvaluatorSuiteStorage[SuiteParameters, CaseParameters]
    | EvaluatorSuiteData[SuiteParameters, CaseParameters]
    | Path
    | str
    | None = None,
    state: Collection[State] = (),
) -> Callable[
    [EvaluatorSuiteDefinition[SuiteParameters, CaseParameters]],
    EvaluatorSuite[SuiteParameters, CaseParameters],
]:
    suite_parameters_type: type[SuiteParameters]
    suite_parameters_value: SuiteParameters | None
    if isinstance(suite_parameters, type):
        suite_parameters_type = suite_parameters
        suite_parameters_value = None

    else:
        suite_parameters_type = type(suite_parameters)
        suite_parameters_value = suite_parameters

    suite_storage: EvaluatorSuiteStorage[SuiteParameters, CaseParameters]
    match storage:
        case None:
            if suite_parameters_value is None:
                try:
                    # try to prepare default value if able
                    suite_parameters_value = suite_parameters_type()

                except Exception:
                    raise RuntimeError(
                        "Failed to create default suite parameters."
                        " Can't define evaluation suite without parameters and storage."
                    ) from None

            suite_storage = _EvaluatorSuiteMemoryStorage[SuiteParameters, CaseParameters](
                data_type=EvaluatorSuiteData[suite_parameters_type, case_parameters],
                parameters=suite_parameters_value,
                cases=(),
            )

        case str() as path_str:
            suite_storage = _EvaluatorSuiteFileStorage[SuiteParameters, CaseParameters](
                path=path_str,
                data_type=EvaluatorSuiteData[suite_parameters, case_parameters],
            )

        case Path() as path:
            suite_storage = _EvaluatorSuiteFileStorage[SuiteParameters, CaseParameters](
                path=path,
                data_type=EvaluatorSuiteData[suite_parameters, case_parameters],
            )

        case EvaluatorSuiteData() as data:
            suite_storage = _EvaluatorSuiteMemoryStorage[SuiteParameters, CaseParameters](
                data_type=EvaluatorSuiteData[suite_parameters, case_parameters],
                parameters=data.parameters,
                cases=data.cases,
            )

        case storage:
            suite_storage = storage

    def wrap(
        definition: EvaluatorSuiteDefinition[SuiteParameters, CaseParameters],
    ) -> EvaluatorSuite[SuiteParameters, CaseParameters]:
        return EvaluatorSuite[SuiteParameters, CaseParameters](
            suite_parameters=suite_parameters_type,
            case_parameters=case_parameters,
            definition=definition,
            storage=suite_storage,
            state=state,
        )

    return wrap


class _EvaluatorSuiteMemoryStorage[SuiteParameters: DataModel, CaseParameters: DataModel]:
    __slots__ = (
        "_cases_store",
        "_data_type",
        "_parameters_store",
    )

    def __init__(
        self,
        data_type: type[EvaluatorSuiteData[SuiteParameters, CaseParameters]],
        parameters: SuiteParameters,
        cases: Sequence[EvaluatorSuiteCase[CaseParameters]] | None,
    ) -> None:
        self._parameters_store: SuiteParameters = parameters
        self._cases_store: Sequence[EvaluatorSuiteCase[CaseParameters]] = cases or ()
        self._data_type: type[EvaluatorSuiteData[SuiteParameters, CaseParameters]] = data_type

    async def load(
        self,
    ) -> EvaluatorSuiteData[SuiteParameters, CaseParameters]:
        return self._data_type(
            parameters=self._parameters_store,
            cases=self._cases_store,
        )

    async def save(
        self,
        data: EvaluatorSuiteData[SuiteParameters, CaseParameters],
    ) -> None:
        self._cases_store = data.cases


class _EvaluatorSuiteFileStorage[SuiteParameters: DataModel, CaseParameters: DataModel]:
    __slots__ = (
        "_data_type",
        "_path",
    )

    def __init__(
        self,
        path: Path | str,
        data_type: type[EvaluatorSuiteData[SuiteParameters, CaseParameters]],
    ) -> None:
        self._path: Path
        match path:
            case str() as path_str:
                self._path = Path(path_str)

            case path:
                self._path = path

        self._data_type: type[EvaluatorSuiteData[SuiteParameters, CaseParameters]] = data_type

    async def load(
        self,
    ) -> EvaluatorSuiteData[SuiteParameters, CaseParameters]:
        try:
            return await self._file_load(data_type=self._data_type)

        except ValueError as exc:
            ctx.log_error(
                f"Invalid EvaluationSuite at {self._path}",
                exception=exc,
            )
            raise exc

    @asynchronous
    def _file_load(
        self,
        data_type: type[EvaluatorSuiteData[SuiteParameters, CaseParameters]],
    ) -> EvaluatorSuiteData[SuiteParameters, CaseParameters]:
        if self._path.exists():
            with open(self._path, mode="rb") as file:
                return data_type.from_json(file.read())

        else:
            raise RuntimeError(f"Missing evaluation suite data at {self._path}")

    async def save(
        self,
        data: EvaluatorSuiteData[SuiteParameters, CaseParameters],
    ) -> None:
        await self._file_save(data=data)

    @asynchronous
    def _file_save(
        self,
        data: EvaluatorSuiteData[SuiteParameters, CaseParameters],
    ) -> None:
        with open(self._path, mode="wb+") as file:
            file.write(data.to_json(indent=2).encode("utf-8"))



from draive.gemini.client import Gemini
from draive.gemini.config import (
    GeminiEmbeddingConfig,
    GeminiGenerationConfig,
    GeminiLiveConfig,
    GeminiSafetyConfig,
    GeminiSafetyThreshold,
)
from draive.gemini.types import GeminiException

__all__ = (
    "Gemini",
    "GeminiEmbeddingConfig",
    "GeminiException",
    "GeminiGenerationConfig",
    "GeminiLiveConfig",
    "GeminiSafetyConfig",
    "GeminiSafetyThreshold",
)



import random
from collections.abc import Mapping, Sequence
from typing import Any
from uuid import UUID, uuid4

from haiway import State, as_dict, ctx, execute_concurrently

from draive.evaluation import EvaluatorSuite, EvaluatorSuiteResult
from draive.evaluation.suite import EvaluatorSuiteCase
from draive.instructions import Instruction, Instructions
from draive.multimodal import MultimodalContent, MultimodalTagElement
from draive.parameters import DataModel
from draive.stages import Stage, StageState, stage

__all__ = ("refine_instruction",)


async def refine_instruction[
    SuiteParameters: DataModel,
    CaseParameters: DataModel,
](
    instruction: Instruction,
    /,
    *,
    guidelines: str | None = None,
    evaluation_suite: EvaluatorSuite[SuiteParameters, CaseParameters],
    rounds_limit: int,
    sample_ratio: float = 0.1,
    candidates_limit: int = 3,
    performance_drop_threshold: float = 0.5,
    quality_threshold: float = 0.99,
    concurrent_nodes: int = 1,
) -> Instruction:
    """
    Refine instruction using binary tree exploration with performance pruning.

    Each node generates exactly 2 complementary strategies.
    Branches are pruned if performance drops below threshold.

    Args:
        instruction: The instruction to refine
        guidelines: Optional guidelines for refinement
        evaluation_suite: Suite to evaluate instruction performance
        rounds_limit: Maximum depth of refinement tree
        sample_ratio: Fraction of passing cases to include in focused evaluation
        candidates_limit: Number of top candidates to fully evaluate
        performance_drop_threshold: Prune branches with score drop > this threshold
        quality_threshold: Stop if score reaches this threshold
        concurrent_nodes: How many nodes explored concurrently
    """

    assert rounds_limit > 0  # nosec: B101
    assert 1 >= sample_ratio > 0  # nosec: B101
    assert candidates_limit > 0  # nosec: B101
    assert 1 >= performance_drop_threshold > 0  # nosec: B101
    assert 1 >= quality_threshold >= 0  # nosec: B101
    assert concurrent_nodes > 0  # nosec: B101

    ctx.log_info(
        f"Starting tree-based refinement: "
        f"max_depth={rounds_limit}, "
        f"sample_ratio={sample_ratio}, "
        f"performance_drop_threshold={performance_drop_threshold}"
    )

    result: MultimodalContent = await Stage.sequence(
        _tree_initialization_stage(
            instruction=instruction,
            evaluation_suite=evaluation_suite,
            sample_ratio=sample_ratio,
            performance_drop_threshold=performance_drop_threshold,
            guidelines=guidelines,
            rounds_limit=rounds_limit,
        ),
        _tree_exploration_stage(
            evaluation_suite=evaluation_suite,
            rounds_limit=rounds_limit,
            quality_threshold=quality_threshold,
            concurrent_nodes=concurrent_nodes,
        ),
        _tree_finalization_stage(
            evaluation_suite=evaluation_suite,
            quality_threshold=quality_threshold,
            candidates_limit=candidates_limit,
        ),
    ).execute()

    ctx.log_info("...instruction refinement finished!")
    return instruction.updated(content=result.to_str())


def _instructions(
    *,
    instruction: Instruction,
) -> Instructions:
    instructions: Instructions = ctx.state(Instructions)

    async def instruction_fetch(
        name: str,
        /,
        *,
        arguments: Mapping[str, str | float | int] | None = None,
        **extra: Any,
    ) -> Instruction | None:
        if name == instruction.name:
            if arguments:
                return instruction.updated(arguments={**instruction.arguments, **arguments})

            else:
                return instruction

        else:  # preserve other prompts unchanged
            return await instructions.fetch(
                name,
                arguments=arguments,
                **extra,
            )

    return Instructions(fetching=instruction_fetch)


class _RefinementTreeNode[SuiteParameters: DataModel, CaseParameters: DataModel](State):
    identifier: UUID
    instruction: Instruction
    strategy: str
    parent_id: UUID | None
    depth: int
    focused_evaluation: EvaluatorSuiteResult[SuiteParameters, CaseParameters]
    complete_evaluation: EvaluatorSuiteResult[SuiteParameters, CaseParameters] | None
    children: Sequence[UUID]
    pruned: bool

    @property
    def is_root(self) -> bool:
        return self.parent_id is None

    @property
    def is_leaf(self) -> bool:
        return len(self.children) == 0

    @property
    def focused_evaluation_performance(self) -> float:
        return self.focused_evaluation.performance

    @property
    def complete_evaluation_performance(self) -> float | None:
        if self.complete_evaluation is None:
            return None

        return self.complete_evaluation.performance


class _RefinementState[
    SuiteParameters: DataModel,
    CaseParameters: DataModel,
](State):
    root: _RefinementTreeNode
    nodes: Mapping[UUID, _RefinementTreeNode[SuiteParameters, CaseParameters]]
    sample_ratio: float
    performance_drop_threshold: float
    guidelines: str | None = None
    rounds_remaining: int

    def leafs(self) -> Sequence[_RefinementTreeNode[SuiteParameters, CaseParameters]]:
        return [node for node in self.nodes.values() if node.is_leaf and not node.pruned]


def _select_focused_cases[
    SuiteParameters: DataModel,
    CaseParameters: DataModel,
](
    *,
    evaluation_result: EvaluatorSuiteResult[SuiteParameters, CaseParameters],
    evaluation_cases: Sequence[EvaluatorSuiteCase[CaseParameters]],
    sample_ratio: float,
) -> Sequence[EvaluatorSuiteCase[CaseParameters]]:
    # Get all failing cases
    failing_cases: Sequence[EvaluatorSuiteCase[CaseParameters]] = [
        case_result.case for case_result in evaluation_result.cases if not case_result.passed
    ]

    # Get passing cases and sample
    passing_cases: Sequence[EvaluatorSuiteCase[CaseParameters]] = [
        case_result.case for case_result in evaluation_result.cases if case_result.passed
    ]

    # Get other, previously excluded cases
    additional_cases: Sequence[EvaluatorSuiteCase[CaseParameters]] = [
        case for case in evaluation_cases if case not in evaluation_result.cases
    ]

    # Intelligent sampling: sample some passing cases
    sampling_cases_pool: Sequence[EvaluatorSuiteCase[CaseParameters]] = (
        passing_cases + additional_cases
    )
    sample_size: int = (
        max(1, int(len(sampling_cases_pool) * sample_ratio)) if sampling_cases_pool else 0
    )
    sampling_cases: Sequence[EvaluatorSuiteCase[CaseParameters]] = (
        random.sample(sampling_cases_pool, min(len(sampling_cases_pool), sample_size))  # nosec: B311
        if sample_size > 0
        else []
    )

    ctx.log_info(
        f"Created focused evaluation suite: "
        f"{len(failing_cases)} failing + {len(sampling_cases)} sampled other "
        f"(out of {len(passing_cases)} total passing)"
    )

    # Combine for focused evaluation
    return failing_cases + sampling_cases


def _tree_initialization_stage[
    SuiteParameters: DataModel,
    CaseParameters: DataModel,
](
    *,
    instruction: Instruction,
    evaluation_suite: EvaluatorSuite[SuiteParameters, CaseParameters],
    sample_ratio: float,
    performance_drop_threshold: float,
    guidelines: str | None,
    rounds_limit: int,
) -> Stage:
    @stage
    async def tree_initialization(
        *,
        state: StageState,
    ) -> StageState:
        ctx.log_info("...evaluating initial instruction with full suite...")
        # Make initial evaluation
        evaluation: EvaluatorSuiteResult[SuiteParameters, CaseParameters]
        with ctx.updated(_instructions(instruction=instruction)):
            evaluation = await evaluation_suite()

        ctx.log_info(f"...initial score: {evaluation.performance:.2f}...")

        # Create root node
        root_node = _RefinementTreeNode(
            identifier=uuid4(),
            instruction=instruction,
            strategy="initial",
            parent_id=None,
            depth=0,
            children=(),
            # for root complete evaluation is the same as focused
            focused_evaluation=evaluation,
            complete_evaluation=evaluation,
            pruned=False,
        )

        # Setup initial state
        return state.updated(
            _RefinementState[SuiteParameters, CaseParameters](
                root=root_node,
                nodes={root_node.identifier: root_node},
                sample_ratio=sample_ratio,
                performance_drop_threshold=performance_drop_threshold,
                guidelines=guidelines,
                rounds_remaining=rounds_limit,
            ),
            # set the result to be initial instruction
            result=MultimodalContent.of(instruction.content),
        )

    return tree_initialization


def _tree_exploration_stage[
    SuiteParameters: DataModel,
    CaseParameters: DataModel,
](
    *,
    evaluation_suite: EvaluatorSuite[SuiteParameters, CaseParameters],
    rounds_limit: int,
    quality_threshold: float,
    concurrent_nodes: int,
) -> Stage:
    @stage
    async def tree_exploration(
        *,
        state: StageState,
    ) -> StageState:
        refinement_state: _RefinementState = state.get(
            _RefinementState,
            required=True,
        )
        evaluation_cases: Sequence[
            EvaluatorSuiteCase[CaseParameters]
        ] = await evaluation_suite.cases()

        async def explore(
            node: _RefinementTreeNode[SuiteParameters, CaseParameters],
        ) -> Mapping[UUID, _RefinementTreeNode[SuiteParameters, CaseParameters]]:
            return await _explore_node(
                node=node,
                evaluation_suite=evaluation_suite,
                evaluation_cases=evaluation_cases,
                sample_ratio=refinement_state.sample_ratio,
                performance_drop_threshold=refinement_state.performance_drop_threshold,
                guidelines=refinement_state.guidelines,
            )

        node_updates: Sequence[
            Mapping[UUID, _RefinementTreeNode[SuiteParameters, CaseParameters]]
        ] = await execute_concurrently(
            explore,
            refinement_state.leafs(),
            concurrent_tasks=concurrent_nodes,
        )

        updated_nodes: dict[UUID, _RefinementTreeNode[SuiteParameters, CaseParameters]] = as_dict(
            refinement_state.nodes
        )
        for nodes in node_updates:
            updated_nodes = {
                **updated_nodes,
                **nodes,
            }

        # Update the refinement state with the updated nodes
        refinement_state = refinement_state.updated(nodes=updated_nodes)

        # Log tree statistics
        total_nodes: int = len(refinement_state.nodes)
        pruned_count: int = len([node for node in refinement_state.nodes.values() if node.pruned])
        active_nodes: int = total_nodes - pruned_count

        ctx.log_info(
            f"Tree exploration round complete. Nodes: {total_nodes} total, "
            f"{active_nodes} active, {pruned_count} pruned"
        )

        return state.updated(refinement_state)

    async def condition(
        *,
        state: StageState,
        iteration: int,
    ) -> bool:
        refinement_state: _RefinementState = state.get(
            _RefinementState,
            required=True,
        )
        if any(
            (
                node.complete_evaluation_performance
                if node.complete_evaluation_performance is not None
                else node.focused_evaluation_performance
            )
            > quality_threshold
            for node in refinement_state.nodes.values()
        ):
            # we can complete early with exceptional score
            return False

        return iteration < rounds_limit

    return Stage.loop(
        tree_exploration,
        condition=condition,
    )


async def _explore_node[
    SuiteParameters: DataModel,
    CaseParameters: DataModel,
](
    node: _RefinementTreeNode[SuiteParameters, CaseParameters],
    evaluation_suite: EvaluatorSuite[SuiteParameters, CaseParameters],
    evaluation_cases: Sequence[EvaluatorSuiteCase[CaseParameters]],
    sample_ratio: float,
    performance_drop_threshold: float,
    guidelines: str | None,
) -> Mapping[UUID, _RefinementTreeNode[SuiteParameters, CaseParameters]]:
    assert not node.pruned  # nosec: B101 # skip pruned branches
    ctx.log_info(
        f"Exploring node {node.identifier} at"
        f" depth {node.depth}, score: {node.focused_evaluation_performance:.2f}"
    )

    # Generate exactly 2 complementary strategies
    strategies: Sequence[tuple[str, Instruction]] = await _generate_refined_instructions(
        instruction=node.instruction,
        evaluation_result=node.focused_evaluation,
        parent_strategy=node.strategy,
        guidelines=guidelines,
    )

    # Create and evaluate child nodes
    focused_suite_cases: Sequence[EvaluatorSuiteCase[CaseParameters]] = _select_focused_cases(
        evaluation_result=node.focused_evaluation,
        evaluation_cases=evaluation_cases,
        sample_ratio=sample_ratio,
    )

    children: dict[UUID, _RefinementTreeNode[SuiteParameters, CaseParameters]] = {}
    for strategy_name, refined_instruction in strategies:
        # Evaluate with focused suite
        ctx.log_info(f"Evaluating strategy '{strategy_name}'...")
        focused_evaluation: EvaluatorSuiteResult[SuiteParameters, CaseParameters]
        with ctx.updated(_instructions(instruction=refined_instruction)):
            focused_evaluation = await evaluation_suite(focused_suite_cases)

        # Check for performance drop
        performance_ratio = (
            focused_evaluation.performance / node.focused_evaluation_performance
            if node.focused_evaluation_performance > 0
            else 0
        )
        pruned: bool = performance_ratio < performance_drop_threshold

        # Create child node
        child_node = _RefinementTreeNode(
            identifier=uuid4(),
            instruction=refined_instruction,
            strategy=strategy_name,
            parent_id=node.identifier,
            depth=node.depth + 1,
            focused_evaluation=focused_evaluation,
            complete_evaluation=None,
            children=(),
            pruned=pruned,
        )
        children[child_node.identifier] = child_node

        if pruned:
            ctx.log_warning(
                f"Strategy '{strategy_name}' caused {(1 - performance_ratio) * 100:.1f}% "
                f"performance drop. Pruning this branch."
            )

        else:
            ctx.log_info(
                f"Created child node {child_node.identifier}: strategy={strategy_name}, "
                f"score={child_node.focused_evaluation_performance:.2f}"
                f" ({performance_ratio:.1%} of parent)"
            )

    return children


async def _generate_strategy_metadata[
    SuiteParameters: DataModel,
    CaseParameters: DataModel,
](
    *,
    evaluation_result: EvaluatorSuiteResult[SuiteParameters, CaseParameters],
    parent_strategy: str | None,
    guidelines: str | None,
) -> Sequence[tuple[str, str]]:
    """
    Generate strategy metadata (name and approach) without instruction content.

    Returns:
        Sequence of tuples containing (name, approach) for each strategy.
    """
    # Analyze failures
    failure_report = evaluation_result.report(
        detailed=True,
        include_passed=False,
    )

    strategy_prompt: str = f"""
You are an expert prompt engineer providing feedback and recemmondations for instruction improvement.
Based on the evaluation failures analysis, prepare EXACTLY 2 DIFFERENT refinement strategies.

<previous_strategy>{parent_strategy or "Initial"}</previous_strategy>
<task>
Generate 2 instruction refinement strategies and recommendations that:
1. Take CONTRASTING approaches to fixing the issues
2. Address different aspects of the failures
3. Are mutually exclusive in their approach
</task>
{f"\n<guidelines>\n{guidelines}</guidelines>\n" if guidelines else ""}
<examples>
<strategy>
<name>Examplification</name>
<approach>Add more specific details and examples</approach>
</strategy>
<strategy>
<name>Constraning</name>
<approach>Add explicit constraints and rules</approach>
</strategy>
</examples>
<format>
For each strategy, provide the name and approach in the following tags format:
<strategy>
<name>[Brief descriptive name]</name>
<approach>[Explanation of the approach with description of recommendations]</approach>
</strategy>
</format>
"""  # noqa: E501

    response: MultimodalContent = await Stage.completion(
        f"<failure_analysis>\n{failure_report}\n</failure_analysis>",
        instruction=strategy_prompt,
    ).execute()

    # Parse strategies
    strategies: list[tuple[str, str]] = []
    for strategy_element in MultimodalTagElement.parse(
        "strategy",
        content=response,
    ):
        name_element: MultimodalTagElement | None = MultimodalTagElement.parse_first(
            "name",
            content=strategy_element.content,
        )
        approach_element: MultimodalTagElement | None = MultimodalTagElement.parse_first(
            "approach",
            content=strategy_element.content,
        )

        if name_element and approach_element:
            strategy_name: str = name_element.content.to_str().strip()
            strategy_approach: str = approach_element.content.to_str().strip()
            strategies.append((strategy_name, strategy_approach))

    return strategies


async def _generate_instruction_content[
    SuiteParameters: DataModel,
    CaseParameters: DataModel,
](
    *,
    current_instruction_content: str,
    strategy_name: str,
    strategy_approach: str,
    evaluation_result: EvaluatorSuiteResult[SuiteParameters, CaseParameters],
    guidelines: str | None,
) -> str:
    # Analyze failures
    failure_report: str = evaluation_result.report(
        detailed=True,
        include_passed=False,
    )

    refinement_prompt: str = f"""\
You are an expert prompt engineer refining provided instructions referred as a subject.
Provide an updated version of the subject that implements the described strategy to address the evaluation failures.
Strictly focus on delivering the best possible instruction content with excellent clarity and performance.
Ensure to apply modifications by following requested strategy while keeping original instruction intent, goal and formatting style.

<strategy>
<name>
{strategy_name}
</name>
<approach>
{strategy_approach}
</approach>
</strategy>

<subject>
{current_instruction_content}
</subject>
{f"\n<guidelines>\n{guidelines}</guidelines>\n" if guidelines else ""}
<format>
Provide ONLY the refined instruction content, without any explanation or metadata.
</format>
"""  # noqa: E501

    ctx.log_info(f"Generating updated instruction using {strategy_name}:\n\n{strategy_approach}")
    response: MultimodalContent = await Stage.completion(
        f"<failure_analysis>\n{failure_report}\n</failure_analysis>",
        instruction=refinement_prompt,
    ).execute()

    updated_instruction: str = response.to_str().strip()

    ctx.log_info(f"Prepared updated instruction using {strategy_name}:\n\n{updated_instruction}")

    return updated_instruction


async def _generate_refined_instructions[
    SuiteParameters: DataModel,
    CaseParameters: DataModel,
](
    instruction: Instruction,
    evaluation_result: EvaluatorSuiteResult[SuiteParameters, CaseParameters],
    parent_strategy: str | None,
    guidelines: str | None,
) -> Sequence[tuple[str, Instruction]]:
    # Step 1: Generate strategy metadata
    strategy_metadata: Sequence[tuple[str, str]] = await _generate_strategy_metadata(
        evaluation_result=evaluation_result,
        parent_strategy=parent_strategy,
        guidelines=guidelines,
    )

    # Step 2: Generate instruction content for each strategy
    refined_strategies: list[tuple[str, Instruction]] = []
    for strategy_name, strategy_approach in strategy_metadata:
        refined_content = await _generate_instruction_content(
            current_instruction_content=instruction.content,
            strategy_name=strategy_name,
            strategy_approach=strategy_approach,
            evaluation_result=evaluation_result,
            guidelines=guidelines,
        )
        refined_instruction: Instruction = instruction.updated(content=refined_content)
        refined_strategies.append((strategy_name, refined_instruction))

    # Validate and return results
    match len(refined_strategies):
        case 2:
            return refined_strategies

        case 1:
            ctx.log_warning(f"Expected 2 strategies, got {len(refined_strategies)}")
            return refined_strategies

        case 0:
            ctx.log_warning("Failed to generate refinement strategies")
            return refined_strategies  # this branch will be skipped

        case _:  # more than 2
            ctx.log_warning(f"Expected 2 strategies, got {len(refined_strategies)}")
            return refined_strategies[:2]


def _tree_finalization_stage[
    SuiteParameters: DataModel,
    CaseParameters: DataModel,
](
    *,
    evaluation_suite: EvaluatorSuite[SuiteParameters, CaseParameters],
    quality_threshold: float,
    candidates_limit: int,
) -> Stage:
    @stage
    async def tree_finalization(
        *,
        state: StageState,
    ) -> StageState:
        refinement_state: _RefinementState = state.get(
            _RefinementState,
            required=True,
        )
        # Find best candidates across entire tree (except root)
        candidates: Sequence[_RefinementTreeNode] = _find_best_candidates(
            refinement_state=refinement_state,
            quality_threshold=quality_threshold,
            limit=candidates_limit,
        )

        if not candidates:
            ctx.log_warning("No candidates found, keeping original instruction")
            return state

        ctx.log_info(f"Running full evaluation on top {len(candidates)} candidates")

        # Full evaluation of top candidates
        best_instruction: Instruction = refinement_state.root.instruction
        best_score: float = refinement_state.root.complete_evaluation_performance or 0
        best_node: _RefinementTreeNode = refinement_state.root

        for candidate_node in candidates:
            ctx.log_info(
                f"Full evaluation of node {candidate_node.identifier} "
                f"(strategy: {candidate_node.strategy}, depth: {candidate_node.depth})"
            )

            complete_evaluation: EvaluatorSuiteResult[SuiteParameters, CaseParameters]
            with ctx.updated(_instructions(instruction=candidate_node.instruction)):
                complete_evaluation = await evaluation_suite()

            # Update node with full eval score
            updated_node: _RefinementTreeNode = candidate_node.updated(
                complete_evaluation=complete_evaluation
            )
            refinement_state = refinement_state.updated(
                nodes={
                    **refinement_state.nodes,
                    # update node in tree
                    updated_node.identifier: updated_node,
                },
            )

            ctx.log_info(
                f"Full evaluation score: {complete_evaluation.performance:.4f} "
                f"(focused was: {updated_node.focused_evaluation_performance:.4f})"
            )

            if complete_evaluation.performance > best_score:
                best_instruction = updated_node.instruction
                best_score = complete_evaluation.performance
                best_node = updated_node

        # Log final statistics
        _log_tree_statistics(
            refinement_state=refinement_state,
            best_node=best_node,
        )

        # Update result with best instruction and updated state
        return state.updated(
            refinement_state,
            result=MultimodalContent.of(
                best_instruction.content,
                meta={"score": best_score},
            ),
        )

    return tree_finalization


def _find_best_candidates(
    *,
    refinement_state: _RefinementState,
    quality_threshold: float,
    limit: int,
) -> Sequence[_RefinementTreeNode]:
    # Get all nodes that are not pruned
    candidates: list[_RefinementTreeNode] = []
    for node in refinement_state.nodes.values():
        # skip excluded and root from candidates
        if node.pruned or node.is_root:
            continue

        # Include if it's a leaf or has exceptional performance
        if node.is_leaf or node.focused_evaluation_performance > quality_threshold:
            candidates.append(node)

    # Sort by focused score and depth (prefer deeper nodes with same score)
    candidates.sort(
        key=lambda x: (x.focused_evaluation_performance, x.depth),
        reverse=True,
    )

    return candidates[:limit]


def _log_tree_statistics(
    *,
    refinement_state: _RefinementState,
    best_node: _RefinementTreeNode,
) -> None:
    total_nodes: int = len(refinement_state.nodes)
    pruned_count: int = len([node for node in refinement_state.nodes.values() if node.pruned])
    active_nodes: int = total_nodes - pruned_count

    # Depth distribution
    depth_distribution: dict[int, int] = {}
    for node in refinement_state.nodes.values():
        depth_distribution[node.depth] = depth_distribution.get(node.depth, 0) + 1

    max_depth: int = max(depth_distribution.keys()) if depth_distribution else 0

    # Build path to best node
    best_path: list[str] = []
    current_id: UUID | None = best_node.identifier
    while current_id:
        node = refinement_state.nodes[current_id]
        best_path.append(f"{node.strategy} (score: {node.focused_evaluation_performance:.3f})")
        current_id = node.parent_id

    best_path.reverse()

    ctx.log_info(
        f"Tree statistics:\n"
        f"- Total nodes explored: {total_nodes}\n"
        f"- Active nodes: {active_nodes}\n"
        f"- Pruned nodes: {pruned_count} ({pruned_count / total_nodes * 100:.1f}%)\n"
        f"- Max depth reached: {max_depth}\n"
        f"- Best path: {' -> '.join(best_path) if best_path else 'None'}"
    )



from collections.abc import Callable, Coroutine
from typing import Any, Protocol, Self, cast, final, overload

from haiway import Meta, MetaValues, ctx
from haiway.utils import format_str

from draive.lmm import LMMToolSpecification
from draive.multimodal import MultimodalContent, MultimodalContentConvertible
from draive.parameters import ParameterSpecification, ParametersSpecification, ParametrizedFunction
from draive.tools.types import (
    ToolAvailabilityChecking,
    ToolError,
    ToolErrorFormatting,
    ToolHandling,
    ToolResultFormatting,
)

__all__ = ("tool",)


@final
class FunctionTool[**Args, Result](ParametrizedFunction[Args, Coroutine[None, None, Result]]):
    __slots__ = (
        "_check_availability",
        "description",
        "format_error",
        "format_result",
        "handling",
        "meta",
        "name",
        "parameters",
        "specification",
    )

    def __init__(
        self,
        /,
        function: Callable[Args, Coroutine[None, None, Result]],
        *,
        name: str,
        description: str | None,
        parameters: ParametersSpecification | None,
        availability: ToolAvailabilityChecking | None,
        format_result: ToolResultFormatting[Result],
        format_error: ToolErrorFormatting,
        handling: ToolHandling = "auto",
        meta: Meta,
    ) -> None:
        super().__init__(function)

        if parameters is None:
            aliased_required: list[str] = []
            specifications: dict[str, ParameterSpecification] = {}
            for parameter in self._parameters.values():
                specifications[parameter.alias or parameter.name] = parameter.specification

                if parameter.required:
                    aliased_required.append(parameter.alias or parameter.name)

            parameters = {
                "type": "object",
                "properties": specifications,
                "required": aliased_required,
            }

        if not parameters["properties"]:
            parameters = None  # use no parameters without arguments

        self.name: str
        object.__setattr__(
            self,
            "name",
            name,
        )
        self.description: str | None
        object.__setattr__(
            self,
            "description",
            description,
        )
        self.parameters: ParametersSpecification | None
        object.__setattr__(
            self,
            "parameters",
            parameters,
        )
        self.specification: LMMToolSpecification
        object.__setattr__(
            self,
            "specification",
            {
                "name": name,
                "description": description,
                "parameters": parameters,
            },
        )
        self.handling: ToolHandling
        object.__setattr__(
            self,
            "handling",
            handling,
        )
        self._check_availability: ToolAvailabilityChecking
        object.__setattr__(
            self,
            "_check_availability",
            availability or _available,  # available by default
        )
        self.format_result: ToolResultFormatting[Result]
        object.__setattr__(
            self,
            "format_result",
            format_result,
        )
        self.format_error: ToolErrorFormatting
        object.__setattr__(
            self,
            "format_error",
            format_error,
        )
        self.meta: Meta
        if description:
            object.__setattr__(
                self,
                "meta",
                meta.updated(
                    kind="tool",
                    name=name,
                    description=description,
                ),
            )

        else:
            object.__setattr__(
                self,
                "meta",
                meta.updated(
                    kind="tool",
                    name=name,
                ),
            )

    def updated(
        self,
        *,
        name: str | None = None,
        description: str | None = None,
        function: Callable[Args, Coroutine[None, None, Result]] | None = None,
        parameters: ParametersSpecification | None = None,
        availability: ToolAvailabilityChecking | None = None,
        format_result: ToolResultFormatting[Result] | None = None,
        format_error: ToolErrorFormatting | None = None,
        handling: ToolHandling | None = None,
        meta: Meta | None = None,
    ) -> Self:
        return self.__class__(
            function or self._call,
            name=name or self.name,
            description=description or self.description,
            parameters=parameters or self.parameters,
            availability=availability or self._check_availability,
            format_result=format_result or self.format_result,
            format_error=format_error or self.format_error,
            handling=handling or self.handling,
            meta=meta or self.meta,
        )

    def available(
        self,
        tools_turn: int,
    ) -> bool:
        try:
            return self._check_availability(
                tools_turn=tools_turn,
                meta=self.meta,
            )

        except Exception as e:
            ctx.log_error(
                f"Availability check of tool ({self.name}) failed, tool will be unavailable.",
                exception=e,
            )
            return False

    # call as a tool
    async def call(
        self,
        call_id: str,
        /,
        **arguments: Any,
    ) -> MultimodalContent:
        async with ctx.scope(f"tool.{self.name}"):
            ctx.record(
                attributes={
                    "call_id": call_id,
                    **{key: f"{arg}" for key, arg in arguments.items()},
                }
            )

            try:
                result: Result = await super().__call__(**arguments)  # pyright: ignore[reportCallIssue]
                formatted_result: MultimodalContent = MultimodalContent.of(
                    self.format_result(result)
                )

                ctx.record(
                    event="result",
                    attributes={"value": format_str(formatted_result)},
                )

                return formatted_result

            except Exception as exc:
                ctx.record(
                    event="error",
                    attributes={
                        "exception": f"{type(exc)}",
                        "message": f"{exc}",
                    },
                )
                error_message: str = f"Tool {self.name} call [{call_id}] failed due to an error"
                ctx.log_error(
                    error_message,
                    exception=exc,
                )
                # raise an error with formatted content
                raise ToolError(
                    error_message,
                    content=MultimodalContent.of(self.format_error(exc)),
                ) from exc

            # do not catch and format BaseExceptions, rethrow instead
            except BaseException as exc:
                ctx.record(
                    event="error",
                    attributes={
                        "exception": f"{type(exc)}",
                        "message": f"{exc}",
                    },
                )
                ctx.log_error(
                    f"Tool {self.name} call [{call_id}] failed due to an error",
                    exception=exc,
                )
                raise exc

    # regular call when using as a function
    async def __call__(
        self,
        *args: Args.args,
        **kwargs: Args.kwargs,
    ) -> Result:
        return await super().__call__(
            *args,
            **kwargs,
        )


def _available(
    tools_turn: int,
    meta: Meta,
) -> bool:
    return True


class ToolWrapper(Protocol):
    def __call__[**Args, Result](
        self,
        function: Callable[Args, Coroutine[None, None, Result]],
    ) -> FunctionTool[Args, Result]: ...


class PartialToolWrapper[Result](Protocol):
    def __call__[**Args](
        self,
        function: Callable[Args, Coroutine[None, None, Result]],
    ) -> FunctionTool[Args, Result]: ...


@overload
def tool[**Args, Result](
    function: Callable[Args, Coroutine[None, None, Result]],
    /,
) -> FunctionTool[Args, Result]:
    """
    Convert a function to a tool using default parameters and no description.

    In order to adjust the arguments behavior and specification use an instance of Argument
    as a default value of any given argument with desired configuration
    for each argument individually.

    Parameters
    ----------
    function: Callable[Args, Coroutine[None, None, Result]]
        a function to be wrapped as a Tool.
    Returns
    -------
    Tool[Args, Result]
        a Tool representation of the provided function.
    """


@overload
def tool[Result](
    *,
    name: str | None = None,
    description: str | None = None,
    availability: ToolAvailabilityChecking | None = None,
    format_result: ToolResultFormatting[Result],
    format_error: ToolErrorFormatting | None = None,
    handling: ToolHandling = "auto",
    meta: Meta | MetaValues | None = None,
) -> PartialToolWrapper[Result]:
    """
    Convert a function to a tool using provided parameters.

    In order to adjust the arguments behavior and specification use an instance of Argument
    as a default value of any given argument with desired configuration
    for each argument individually.

    Parameters
    ----------
    name: str
        name to be used in a tool specification.
        Default is the name of the wrapped function.
    description: int
        description to be used in a tool specification. Allows to present the tool behavior to the
        external system.
        Default is empty.
    availability: ToolAvailabilityChecking
        function used to verify availability of the tool in given context. It can be used to check
        permissions or occurrence of a specific state to allow its usage.
        Provided function should raise an Exception when the tool should not be available.
        Default is always available.
    format_result: Callable[[Result], MultimodalContent]
        function converting tool result to MultimodalContent. It is used to format the result
        for model processing. Default implementation converts the result to string if needed.
    format_error: Callable[[Exception], MultimodalContent]
        function converting tool call exception to a fallback MultimodalContent.
        Default implementation return "ERROR" string and logs the exception.
    handling: ToolHandling
        controls if tool result should break the ongoing processing and be the direct result of it.
        Note that during concurrent execution of multiple tools the call/result order defines
        direct result and exact behavior is not defined.
        Default is "auto".
    meta: Mapping[str, str | float | int | bool | None] | None
        custom metadata allowing to access tool metadata like its source in case of remote tools.

    Returns
    -------
    Callable[[Function[Args, Coroutine[None, None, Result]]], Tool[Args, Result]]
        function allowing to convert other function to a Tool using provided configuration.
    """


@overload
def tool(
    *,
    name: str | None = None,
    description: str | None = None,
    availability: ToolAvailabilityChecking | None = None,
    format_error: ToolErrorFormatting | None = None,
    handling: ToolHandling = "auto",
    meta: Meta | MetaValues | None = None,
) -> ToolWrapper:
    """
    Convert a function to a tool using provided parameters.

    In order to adjust the arguments behavior and specification use an instance of Argument
    as a default value of any given argument with desired configuration
    for each argument individually.

    Parameters
    ----------
    name: str
        name to be used in a tool specification.
        Default is the name of the wrapped function.
    description: int
        description to be used in a tool specification. Allows to present the tool behavior to the
        external system.
        Default is empty.
    availability: ToolAvailabilityChecking
        function used to verify availability of the tool in given context. It can be used to check
        permissions or occurrence of a specific state to allow its usage.
        Provided function should raise an Exception when the tool should not be available.
        Default is always available.
    format_error: Callable[[Exception], MultimodalContent]
        function converting tool call exception to a fallback MultimodalContent.
        Default implementation return "ERROR" string and logs the exception.
    handling: ToolHandling
        controls if tool result should break the ongoing processing and be the direct result of it.
        Note that during concurrent execution of multiple tools the call/result order defines
        direct result and exact behavior is not defined.
        Default is "auto".
    meta: Meta | MetaValues | None
        custom metadata allowing to access tool metadata like its source in case of remote tools.

    Returns
    -------
    Callable[[Function[Args, Coroutine[None, None, Result]]], Tool[Args, Result]]
        function allowing to convert other function to a Tool using provided configuration.
    """


def tool[**Args, Result](
    function: Callable[Args, Coroutine[None, None, Result]] | None = None,
    *,
    name: str | None = None,
    description: str | None = None,
    availability: ToolAvailabilityChecking | None = None,
    format_result: ToolResultFormatting[Result] | None = None,
    format_error: ToolErrorFormatting | None = None,
    handling: ToolHandling = "auto",
    meta: Meta | MetaValues | None = None,
) -> PartialToolWrapper[Result] | ToolWrapper | FunctionTool[Args, Result]:
    def wrap[**Arg](
        function: Callable[Arg, Coroutine[None, None, Result]],
    ) -> FunctionTool[Arg, Result]:
        return FunctionTool[Arg, Result](
            function=function,
            name=name or function.__name__,
            description=description,
            parameters=None,
            availability=availability,
            format_result=format_result or _default_result_format,
            format_error=format_error or _default_error_format,
            handling=handling,
            meta=Meta.of(meta),
        )

    if function := function:
        return wrap(function=function)

    else:
        return cast(PartialToolWrapper[Result], wrap)


def _default_result_format(result: Any) -> MultimodalContent:
    match result:
        case MultimodalContent() as content:
            return content

        case element if isinstance(element, MultimodalContentConvertible):
            return MultimodalContent.of(element)

        case [*elements]:
            return MultimodalContent.of(
                *[
                    element if isinstance(element, MultimodalContentConvertible) else str(element)
                    for element in elements
                ]
            )

        case other:
            return MultimodalContent.of(str(other))


def _default_error_format(error: Exception) -> MultimodalContent:
    return MultimodalContent.of("ERROR")



from collections.abc import Iterable, Sequence
from itertools import chain
from typing import Any

from haiway import ctx

from draive.choice.types import ChoiceOption, SelectionException
from draive.instructions import Instruction
from draive.lmm import (
    LMM,
    LMMCompletion,
    LMMContextElement,
    LMMInput,
    LMMInstruction,
    LMMToolRequests,
    LMMToolResponses,
)
from draive.multimodal import Multimodal, MultimodalContent, MultimodalTagElement
from draive.tools import Toolbox

__all__ = ("default_choice_completion",)


async def default_choice_completion(
    *,
    instruction: Instruction | str,
    options: Sequence[ChoiceOption],
    input: Multimodal,  # noqa: A002
    toolbox: Toolbox,
    examples: Iterable[tuple[Multimodal, ChoiceOption]] | None,
    **extra: Any,
) -> ChoiceOption:
    async with ctx.scope("choice_completion"):
        assert options, "Choice options cannot be empty"  # nosec: B101
        assert all(  # nosec: B101
            example[1] in options for example in examples or []
        ), "Choice options examples have to be included in available options"

        if len(options) == 1:
            return options[0]  # if there is ony one option it has to be the result

        options_map: dict[str, ChoiceOption] = {option.identifier: option for option in options}
        assert len(options_map) == len(options), "Choice options cannot have duplicate identifiers"  # nosec: B101

        extended_instruction: Instruction
        match instruction:
            case str() as instruction_string:
                extended_instruction = Instruction.of(
                    f"{instruction_string}\n\n{INSTRUCTION_EXTENSION}"
                )

            case instruction:
                extended_instruction = instruction.extended(
                    INSTRUCTION_EXTENSION,
                    joiner="\n\n",
                )

        formatted_options: MultimodalContent = MultimodalContent.of(
            *(_format_option(option) for option in options)
        )

        context: list[LMMContextElement] = [
            *chain.from_iterable(
                _format_example(
                    example,
                    options=formatted_options,
                )
                for example in examples or []
            ),
            _format_input(
                input,
                options=formatted_options,
            ),
        ]

        formatted_instruction: LMMInstruction = Instruction.formatted(extended_instruction)
        tools_turn: int = 0
        result: MultimodalContent = MultimodalContent.empty
        result_extension: MultimodalContent = MultimodalContent.empty
        while True:
            ctx.log_debug("...requesting completion...")
            match await LMM.completion(
                instruction=formatted_instruction,
                context=context,
                tools=toolbox.available_tools(tools_turn=tools_turn),
                output="text",
                **extra,
            ):
                case LMMCompletion() as completion:
                    ctx.log_debug("...received result...")
                    result = result_extension.appending(completion.content)
                    break  # proceed to resolving

                case LMMToolRequests() as tool_requests:
                    ctx.log_debug(f"...received tool requests (turn {tools_turn})...")
                    # skip tool_requests.content - no need for extra comments
                    tool_responses: LMMToolResponses = await toolbox.respond_all(tool_requests)

                    if completion := tool_responses.completion(extension=result_extension):
                        ctx.log_debug("...received tools direct result...")
                        result = completion.content
                        break  # proceed to resolving

                    elif extension := tool_responses.completion_extension():
                        ctx.log_debug("...received tools result extension...")
                        result_extension = result_extension.appending(extension)

                    ctx.log_debug("...received tools responses...")
                    context.extend((tool_requests, tool_responses))

            tools_turn += 1  # continue with next turn

        ctx.log_debug("...decoding result...")
        if selection := MultimodalTagElement.parse_first(
            "CHOICE",
            content=result,
        ):
            if option := options_map.get(selection.content.to_str()):
                ctx.log_debug(f"...selection completed with {option.identifier}!")
                return option

        ctx.log_debug("...selection invalid!")
        raise SelectionException("Invalid or missing selection")


def _format_option(
    option: ChoiceOption,
    /,
) -> MultimodalContent:
    return MultimodalContent.of(
        f"<OPTION identifier='{option.identifier}'>",
        option.content,
        "</OPTION>",
    )


def _format_input(
    input: Multimodal,  # noqa: A002
    /,
    options: MultimodalContent,
) -> LMMInput:
    return LMMInput.of(
        MultimodalContent.of(
            "<INPUT>",
            input,
            "</INPUT>\n\n<OPTIONS>",
            options,
            "</OPTIONS>",
        )
    )


def _format_example(
    example: tuple[Multimodal, ChoiceOption],
    /,
    options: MultimodalContent,
) -> tuple[LMMInput, LMMCompletion]:
    return (
        _format_input(example[0], options=options),
        LMMCompletion.of(f"```\n{example[1].identifier}\n```"),
    )


INSTRUCTION_EXTENSION: str = """\
<FORMAT>
Ensure the result is following the strict format:
Place only the identifier of the chosen option inside a <CHOICE> tag, as shown here:

<CHOICE>option_identifier</CHOICE>

</FORMAT>
"""



from collections.abc import (
    AsyncIterator,
    Iterable,
    Mapping,
    Sequence,
)
from types import TracebackType
from typing import (
    Any,
    ClassVar,
    Literal,
    Protocol,
    Self,
    TypedDict,
    overload,
    runtime_checkable,
)

from haiway import META_EMPTY, Default, Meta, MetaValues, State, as_tuple, ctx

from draive.multimodal import Multimodal, MultimodalContent
from draive.parameters import DataModel, ParametersSpecification
from draive.utils import Memory

__all__ = (
    "LMMCompleting",
    "LMMCompletion",
    "LMMContext",
    "LMMContextElement",
    "LMMException",
    "LMMInput",
    "LMMInstruction",
    "LMMMemory",
    "LMMOutput",
    "LMMOutputDecoder",
    "LMMOutputInvalid",
    "LMMOutputLimit",
    "LMMSession",
    "LMMSessionClosing",
    "LMMSessionEvent",
    "LMMSessionInput",
    "LMMSessionOpening",
    "LMMSessionOutput",
    "LMMSessionOutputSelection",
    "LMMSessionReading",
    "LMMSessionScope",
    "LMMSessionWriting",
    "LMMStreamChunk",
    "LMMStreamInput",
    "LMMStreamOutput",
    "LMMToolFunctionSpecification",
    "LMMToolRequest",
    "LMMToolRequests",
    "LMMToolResponse",
    "LMMToolResponseHandling",
    "LMMToolResponses",
    "LMMToolSelection",
    "LMMToolSpecification",
    "LMMTools",
)

LMMInstruction = str


class LMMToolFunctionSpecification(TypedDict):
    name: str
    description: str | None
    parameters: ParametersSpecification | None


type LMMToolSpecification = LMMToolFunctionSpecification
type LMMOutputSelection = (
    Sequence[Literal["text", "image", "audio", "video"]]
    | Literal["auto", "text", "json", "image", "audio", "video"]
    | type[DataModel]
)
type LMMToolSelection = Literal["auto", "required", "none"] | LMMToolSpecification


class LMMTools(State):
    none: ClassVar[Self]  # defined after the class

    @classmethod
    def of(
        cls,
        tools: Iterable[LMMToolSpecification] | None,
        /,
        *,
        selection: LMMToolSelection = "auto",
    ) -> Self:
        specifications: tuple[LMMToolSpecification, ...] | None = as_tuple(tools)
        if not specifications:
            return cls(
                selection="none",
                specifications=(),
            )

        if not isinstance(selection, str) and selection not in specifications:
            return cls(
                selection="auto",
                specifications=specifications,
            )

        return cls(
            selection=selection,
            specifications=specifications,
        )

    selection: LMMToolSelection
    specifications: Sequence[LMMToolSpecification]

    def __bool__(self) -> bool:
        return bool(self.specifications)


LMMTools.none = LMMTools.of((), selection="none")


class LMMException(Exception):
    pass


class LMMOutputLimit(LMMException):
    pass


class LMMOutputInvalid(LMMException):
    pass


class LMMInput(DataModel):
    @classmethod
    def of(
        cls,
        content: Multimodal,
        /,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls(
            content=MultimodalContent.of(content),
            meta=Meta.of(meta),
        )

    content: MultimodalContent
    meta: Meta = META_EMPTY

    def __bool__(self) -> bool:
        return bool(self.content)


class LMMCompletion(DataModel):
    @classmethod
    def of(
        cls,
        content: Multimodal,
        /,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls(
            content=MultimodalContent.of(content),
            meta=Meta.of(meta),
        )

    content: MultimodalContent
    meta: Meta = META_EMPTY

    def __bool__(self) -> bool:
        return bool(self.content)


@runtime_checkable
class LMMOutputDecoder(Protocol):
    def __call__(
        self,
        content: MultimodalContent,
    ) -> MultimodalContent: ...


LMMToolResponseHandling = Literal["result", "error", "completion", "extension", "detached"]


class LMMToolResponse(DataModel):
    @classmethod
    def of(
        cls,
        identifier: str,
        /,
        *,
        tool: str,
        content: MultimodalContent,
        handling: LMMToolResponseHandling,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls(
            identifier=identifier,
            tool=tool,
            content=content,
            handling=handling,
            meta=Meta.of(meta),
        )

    identifier: str
    tool: str
    content: MultimodalContent
    handling: LMMToolResponseHandling
    meta: Meta = META_EMPTY


class LMMToolResponses(DataModel):
    @classmethod
    def of(
        cls,
        responses: Sequence[LMMToolResponse],
        /,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls(
            responses=responses,
            meta=Meta.of(meta),
        )

    responses: Sequence[LMMToolResponse]
    meta: Meta = META_EMPTY

    def completion(
        self,
        *,
        extension: MultimodalContent | None = None,
    ) -> LMMCompletion | None:
        completion_parts: list[MultimodalContent] = []
        extension_parts: list[MultimodalContent] = [extension] if extension else []
        merged_meta: Meta = self.meta
        for response in self.responses:
            if response.handling == "extension":
                extension_parts.append(response.content)
                continue  # include extensions

            if response.handling != "completion":
                continue  # collect completion

            completion_parts.append(response.content)
            merged_meta = merged_meta.merged_with(response.meta)

        if not completion_parts:
            return None

        return LMMCompletion.of(
            MultimodalContent.of(*extension_parts, *completion_parts),
            meta=merged_meta,
        )

    def completion_extension(self) -> MultimodalContent:
        return MultimodalContent.of(
            *(response.content for response in self.responses if response.handling == "extension")
        )


class LMMToolRequest(DataModel):
    @classmethod
    def of(
        cls,
        identifier: str,
        /,
        *,
        tool: str,
        arguments: Mapping[str, Any] | None = None,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls(
            identifier=identifier,
            tool=tool,
            arguments=arguments if arguments is not None else {},
            meta=Meta.of(meta),
        )

    identifier: str
    tool: str
    arguments: Mapping[str, Any] = Default(factory=dict)
    meta: Meta = META_EMPTY


class LMMToolRequests(DataModel):
    @classmethod
    def of(
        cls,
        requests: Sequence[LMMToolRequest],
        /,
        content: MultimodalContent | None = None,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls(
            requests=requests,
            content=content if content is not None else MultimodalContent.empty,
            meta=Meta.of(meta),
        )

    requests: Sequence[LMMToolRequest]
    content: MultimodalContent = MultimodalContent.empty
    meta: Meta = META_EMPTY


LMMContextElement = LMMInput | LMMCompletion | LMMToolRequests | LMMToolResponses
LMMContext = Sequence[LMMContextElement]
LMMOutput = LMMCompletion | LMMToolRequests


class LMMStreamChunk(DataModel):
    @classmethod
    def of(
        cls,
        content: Multimodal,
        /,
        eod: bool = False,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls(
            content=MultimodalContent.of(content),
            eod=eod,
            meta=Meta.of(meta),
        )

    content: MultimodalContent
    eod: bool
    meta: Meta = META_EMPTY

    def __bool__(self) -> bool:
        return bool(self.content)


LMMStreamInput = LMMStreamChunk | LMMToolResponse
LMMStreamOutput = LMMStreamChunk | LMMToolRequest
LMMMemory = Memory[LMMContext, LMMContextElement]


@runtime_checkable
class LMMCompleting(Protocol):
    @overload
    async def __call__(
        self,
        *,
        instruction: LMMInstruction | None,
        context: LMMContext,
        tools: LMMTools | None,
        output: LMMOutputSelection,
        stream: Literal[False] = False,
        **extra: Any,
    ) -> LMMOutput: ...

    @overload
    async def __call__(
        self,
        *,
        instruction: LMMInstruction | None,
        context: LMMContext,
        tools: LMMTools | None,
        output: LMMOutputSelection,
        stream: Literal[True],
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput]: ...

    async def __call__(
        self,
        *,
        instruction: LMMInstruction | None,
        context: LMMContext,
        tools: LMMTools | None,
        output: LMMOutputSelection,
        stream: bool = False,
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput] | LMMOutput: ...


class LMMSessionEvent(DataModel):
    @classmethod
    def of(
        cls,
        category: str,
        /,
        *,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        return cls(
            category=category,
            meta=Meta.of(meta),
        )

    category: str
    meta: Meta


LMMSessionInput = LMMStreamChunk | LMMToolResponse | LMMSessionEvent
LMMSessionOutput = LMMStreamChunk | LMMToolRequest | LMMSessionEvent
type LMMSessionOutputSelection = (
    Sequence[Literal["text", "audio"]] | Literal["auto", "text", "audio"]
)


@runtime_checkable
class LMMSessionReading(Protocol):
    async def __call__(self) -> LMMSessionOutput: ...


@runtime_checkable
class LMMSessionWriting(Protocol):
    async def __call__(
        self,
        input: LMMSessionInput,  # noqa: A002
    ) -> None: ...


class LMMSession(State):
    @classmethod
    async def read(cls) -> LMMSessionOutput:
        return await ctx.state(cls).reading()

    @classmethod
    async def reader(cls) -> AsyncIterator[LMMSessionOutput]:
        session: Self = ctx.state(cls)
        while True:  # breaks on exception
            yield await session.reading()

    @classmethod
    async def write(
        cls,
        input: LMMSessionInput,  # noqa: A002
    ) -> None:
        await ctx.state(cls).writing(input=input)

    @classmethod
    async def writer(
        cls,
        input: AsyncIterator[LMMSessionInput],  # noqa: A002
    ) -> None:
        session: Self = ctx.state(cls)
        while True:  # breaks on exception
            await session.writing(input=await anext(input))

    reading: LMMSessionReading
    writing: LMMSessionWriting


@runtime_checkable
class LMMSessionOpening(Protocol):
    async def __call__(self) -> LMMSession: ...


@runtime_checkable
class LMMSessionClosing(Protocol):
    async def __call__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None: ...


class LMMSessionScope(State):
    opening: LMMSessionOpening
    closing: LMMSessionClosing

    async def __aenter__(self) -> LMMSession:
        return await self.opening()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.closing(
            exc_type=exc_type,
            exc_val=exc_val,
            exc_tb=exc_tb,
        )


@runtime_checkable
class LMMSessionPreparing(Protocol):
    async def __call__(
        self,
        *,
        instruction: LMMInstruction | None,
        memory: LMMMemory | None,
        tools: LMMTools | None,
        output: LMMSessionOutputSelection,
        **extra: Any,
    ) -> LMMSessionScope: ...



from collections.abc import AsyncIterator
from typing import Any, Literal, final, overload

from haiway import State, ctx

from draive.lmm.types import (
    LMMCompleting,
    LMMContext,
    LMMInstruction,
    LMMMemory,
    LMMOutput,
    LMMOutputSelection,
    LMMSessionOutputSelection,
    LMMSessionPreparing,
    LMMSessionScope,
    LMMStreamOutput,
    LMMTools,
)

__all__ = (
    "LMM",
    "RealtimeLMM",
)


@final
class LMM(State):
    @overload
    @classmethod
    async def completion(
        cls,
        *,
        instruction: LMMInstruction | None = None,
        context: LMMContext,
        tools: LMMTools | None = None,
        output: LMMOutputSelection = "auto",
        stream: Literal[False] = False,
        **extra: Any,
    ) -> LMMOutput: ...

    @overload
    @classmethod
    async def completion(
        cls,
        *,
        instruction: LMMInstruction | None = None,
        context: LMMContext,
        tools: LMMTools | None = None,
        output: LMMOutputSelection = "auto",
        stream: Literal[True],
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput]: ...

    @classmethod
    async def completion(
        cls,
        *,
        instruction: LMMInstruction | None = None,
        context: LMMContext,
        tools: LMMTools | None = None,
        output: LMMOutputSelection = "auto",
        stream: bool = False,
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput] | LMMOutput:
        if stream:
            return await ctx.state(cls).completing(
                instruction=instruction,
                context=context,
                tools=tools,
                output=output,
                stream=True,
                **extra,
            )

        else:
            return await ctx.state(cls).completing(
                instruction=instruction,
                context=context,
                tools=tools,
                output=output,
                **extra,
            )

    completing: LMMCompleting


@final
class RealtimeLMM(State):
    @classmethod
    async def session(
        cls,
        *,
        instruction: LMMInstruction | None = None,
        memory: LMMMemory | None = None,
        tools: LMMTools | None = None,
        output: LMMSessionOutputSelection = "auto",
        **extra: Any,
    ) -> LMMSessionScope:
        return await ctx.state(cls).session_preparing(
            instruction=instruction,
            memory=memory,
            output=output,
            tools=tools,
            **extra,
        )

    session_preparing: LMMSessionPreparing



from draive.lmm.types import LMMOutputDecoder, LMMOutputInvalid, LMMOutputSelection
from draive.multimodal import MultimodalContent
from draive.parameters.model import DataModel

__all__ = ("lmm_output_decoder",)


def lmm_output_decoder(  # noqa: PLR0911
    output: LMMOutputSelection,
    /,
) -> LMMOutputDecoder:
    match output:
        case "auto":
            return _auto_output_conversion

        case "text":
            return _text_output_conversion

        case "json":
            return _json_output_conversion

        case "image":
            return _image_output_conversion

        case "audio":
            return _audio_output_conversion

        case "video":
            return _video_output_conversion

        case [*_]:  # we could prepare more specific method
            return _auto_output_conversion

        case model:
            return _prepare_model_output_conversion(model)


def _auto_output_conversion(
    content: MultimodalContent,
) -> MultimodalContent:
    return content


def _text_output_conversion(
    content: MultimodalContent,
) -> MultimodalContent:
    return content.text()


def _image_output_conversion(
    content: MultimodalContent,
) -> MultimodalContent:
    return content.images()


def _audio_output_conversion(
    content: MultimodalContent,
) -> MultimodalContent:
    return content.audio()


def _video_output_conversion(
    content: MultimodalContent,
) -> MultimodalContent:
    return content.video()


def _json_output_conversion(
    content: MultimodalContent,
) -> MultimodalContent:
    try:
        return MultimodalContent.of(DataModel.from_json(content.text().to_str()))

    except ValueError as exc:
        raise LMMOutputInvalid("Failed to decode JSON output", content) from exc


def _prepare_model_output_conversion(
    model: type[DataModel],
) -> LMMOutputDecoder:
    def _model_output_conversion(
        content: MultimodalContent,
    ) -> MultimodalContent:
        try:
            return MultimodalContent.of(model.from_json(content.text().to_str()))

        except ValueError as exc:
            raise LMMOutputInvalid("Failed to decode JSON output", content) from exc

    return _model_output_conversion



from collections.abc import Iterable
from typing import Any, Literal, cast

from haiway import ctx

from draive.generation.model.types import ModelGeneratorDecoder
from draive.instructions import Instruction
from draive.lmm import (
    LMM,
    LMMCompletion,
    LMMContextElement,
    LMMInput,
    LMMInstruction,
    LMMToolRequests,
    LMMToolResponses,
)
from draive.multimodal import (
    MultimodalContent,
)
from draive.parameters import DataModel
from draive.prompts import Prompt
from draive.tools import Toolbox

__all__ = ("generate_model",)


async def generate_model[Generated: DataModel](  # noqa: C901, PLR0912, PLR0915
    generated: type[Generated],
    /,
    *,
    instruction: Instruction,
    input: Prompt | MultimodalContent,  # noqa: A002
    schema_injection: Literal["auto", "full", "simplified", "skip"],
    toolbox: Toolbox,
    examples: Iterable[tuple[MultimodalContent, Generated]] | None,
    decoder: ModelGeneratorDecoder | None,
    **extra: Any,
) -> Generated:
    async with ctx.scope("generate_model"):
        extended_instruction: Instruction
        match schema_injection:
            case "auto":
                extended_instruction = instruction.extended(
                    DEFAULT_INSTRUCTION_EXTENSION.format(
                        schema=generated.simplified_schema(indent=2),
                    ),
                    joiner="\n\n",
                )

            case "full":
                extended_instruction = instruction.with_arguments(
                    schema=generated.json_schema(indent=2),
                )

            case "simplified":
                extended_instruction = instruction.with_arguments(
                    schema=generated.simplified_schema(indent=2),
                )

            case "skip":
                extended_instruction = instruction

        context: list[LMMContextElement] = [
            *[
                message
                for example in examples or []
                for message in [
                    LMMInput.of(example[0]),
                    LMMCompletion.of(example[1].to_json(indent=2)),
                ]
            ],
        ]

        match input:
            case Prompt() as prompt:
                context.extend(prompt.content)

            case value:
                context.append(LMMInput.of(value))

        formatted_instruction: LMMInstruction = Instruction.formatted(extended_instruction)
        tools_turn: int = 0
        result: MultimodalContent = MultimodalContent.empty
        result_extension: MultimodalContent = MultimodalContent.empty
        while True:
            ctx.log_debug("...requesting completion...")
            match await LMM.completion(
                instruction=formatted_instruction,
                context=context,
                tools=toolbox.available_tools(tools_turn=tools_turn),
                output=generated,  # provide model specification
                **extra,
            ):
                case LMMCompletion() as completion:
                    ctx.log_debug("...received result...")
                    result = result_extension.appending(completion.content)
                    break  # proceed to resolving

                case LMMToolRequests() as tool_requests:
                    ctx.log_debug(f"...received tool requests (turn {tools_turn})...")
                    # skip tool_requests.content - no need for extra comments
                    tool_responses: LMMToolResponses = await toolbox.respond_all(tool_requests)

                    if completion := tool_responses.completion(extension=result_extension):
                        ctx.log_debug("...received tools direct result...")
                        result = completion.content
                        break  # proceed to resolving

                    elif extension := tool_responses.completion_extension():
                        ctx.log_debug("...received tools result extension...")
                        result_extension = result_extension.appending(extension)

                    ctx.log_debug("...received tools responses...")
                    context.extend((tool_requests, tool_responses))

            tools_turn += 1  # continue with next turn

        try:
            ctx.log_debug("...decoding result...")
            if decoder := decoder:
                ctx.log_debug("...decoded with custom decoder!")
                return decoder(result)

            elif artifacts := result.artifacts(generated):
                ctx.log_debug("...direct artifact found!")
                return cast(Generated, artifacts[0])

            else:
                ctx.log_debug("...decoded from json string!")
                return generated.from_json(result.to_str())

        except Exception as exc:
            ctx.log_error(
                f"Failed to decode {generated.__name__} model due to an error: {type(exc)}",
                exception=exc,
            )
            raise exc


DEFAULT_INSTRUCTION_EXTENSION: str = """\
<FORMAT>
Provide the result using a single raw valid JSON object that adheres strictly to the given \
SCHEMA without any comments, formatting, or additional elements.
<SCHEMA>
{schema}
</SCHEMA>
</FORMAT>
"""



from asyncio import CancelledError
from collections.abc import Sequence
from types import TracebackType
from typing import Any
from uuid import uuid4

from haiway import ctx

from draive.conversation.realtime.types import (
    RealtimeConversationSession,
    RealtimeConversationSessionScope,
)
from draive.conversation.types import (
    ConversationEvent,
    ConversationMemory,
    ConversationMessage,
    ConversationMessageChunk,
    ConversationStreamElement,
)
from draive.instructions import Instruction
from draive.lmm import (
    LMMContext,
    LMMContextElement,
    LMMMemory,
    LMMSession,
    LMMSessionEvent,
    LMMSessionScope,
    LMMStreamChunk,
    LMMToolRequest,
    LMMToolResponse,
    RealtimeLMM,
)
from draive.lmm.types import LMMCompletion, LMMInput
from draive.tools import Toolbox
from draive.utils import MEMORY_NONE, Memory, MemoryRecalling, MemoryRemembering

__all__ = ("realtime_conversation_preparing",)


async def realtime_conversation_preparing(  # noqa: C901
    *,
    instruction: Instruction | None,
    memory: ConversationMemory | None,
    toolbox: Toolbox,
    **extra: Any,
) -> RealtimeConversationSessionScope:
    session_memory: LMMMemory
    match memory:
        case None:
            session_memory = MEMORY_NONE

        case Memory():
            session_memory = Memory(
                recall=_to_lmm_recall(memory.recall),
                remember=_to_lmm_remember(memory.remember),
            )

    # TODO: rework RealtimeLMM interface to allow instruction and tools changes during the session
    session_scope: LMMSessionScope = await RealtimeLMM.session(
        instruction=Instruction.formatted(instruction),
        memory=session_memory,
        tools=toolbox.available_tools(),
        output="auto",
    )

    async def open_session() -> RealtimeConversationSession:  # noqa: C901
        session: LMMSession = await session_scope.__aenter__()

        async def handle_tool_request(
            tool_request: LMMToolRequest,
            /,
        ) -> None:
            try:
                ctx.log_debug(
                    f"Handling tool ({tool_request.tool}) request ({tool_request.identifier}) ...",
                )
                response: LMMToolResponse = await toolbox.respond(tool_request)
                if response.handling in ("completion", "extension"):
                    ctx.log_warning(
                        f"Tool handling `{response.handling}` is not supported in"
                        " realtime conversation, using regular result handling instead"
                    )

                # check if task was not cancelled before passing the result to input
                ctx.check_cancellation()
                # deliver the result directly to input
                await session.writing(response)

            except CancelledError:
                ctx.log_debug(
                    f"...tool request ({tool_request.identifier}) handling cancelled!",
                )

            except BaseException as exc:
                ctx.log_error(
                    f"...tool request ({tool_request.identifier}) handling failed!",
                    exception=exc,
                )

            else:
                ctx.log_debug(
                    f"...tool request ({tool_request.identifier}) handling completed!",
                )

        async def read() -> ConversationStreamElement:
            while True:
                match await session.reading():
                    case LMMStreamChunk() as chunk:
                        ctx.log_debug("...received response chunk...")
                        return ConversationMessageChunk.model(
                            chunk.content,
                            identifier=chunk.meta.identifier,
                            message_identifier=chunk.meta.get_uuid(
                                "message_identifier",
                                default=uuid4(),
                            ),
                            eod=chunk.eod,
                        )

                    case LMMSessionEvent() as event:
                        ctx.log_debug(f"...received {event.category} event...")
                        return ConversationEvent.of(
                            event.category,
                            meta=event.meta,
                        )

                    case LMMToolRequest() as tool_request:
                        ctx.log_debug(f"...received {tool_request.tool} request...")
                        ctx.spawn(
                            handle_tool_request,
                            tool_request,
                        )
                        continue  # keep receiving elements

        async def write(
            input: ConversationStreamElement,  # noqa: A002
        ) -> None:
            if isinstance(input, ConversationMessageChunk):
                # pass chunk to LMM
                await session.writing(
                    input=LMMStreamChunk.of(
                        input.content,
                        eod=input.eod,
                        meta=input.meta,
                    )
                )

            else:
                assert isinstance(input, ConversationEvent)  # nosec: B101
                # pass event to LMM
                await session.writing(
                    input=LMMSessionEvent.of(
                        input.category,
                        meta=input.meta,
                    )
                )

        return RealtimeConversationSession(
            reading=read,
            writing=write,
        )

    async def close_session(
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await session_scope.__aexit__(
            exc_type,
            exc_val,
            exc_tb,
        )

    return RealtimeConversationSessionScope(
        opening=open_session,
        closing=close_session,
    )


def _to_lmm_recall(
    recall: MemoryRecalling[Sequence[ConversationMessage]],
    /,
) -> MemoryRecalling[LMMContext]:
    async def lmm_recall(
        **extra: Any,
    ) -> LMMContext:
        return tuple(element.to_lmm_context() for element in await recall())

    return lmm_recall


def _to_lmm_remember(
    remember: MemoryRemembering[ConversationMessage],
    /,
) -> MemoryRemembering[LMMContextElement]:
    async def lmm_remember(
        *items: LMMContextElement,
        **extra: Any,
    ) -> None:
        await remember(
            *(
                ConversationMessage.from_lmm_context(element)
                for element in items
                if isinstance(element, LMMInput | LMMCompletion)
            )
        )

    return lmm_remember



import random
from collections.abc import AsyncGenerator, AsyncIterator, Callable
from itertools import chain
from typing import Any, Literal, cast, overload

from google.api_core.exceptions import ResourceExhausted  # pyright: ignore[reportMissingImport]
from google.genai.types import (
    Candidate,
    ContentListUnionDict,
    FinishReason,
    FunctionCallingConfigMode,
    FunctionDeclarationDict,
    GenerateContentConfigDict,
    GenerateContentResponse,
    GenerateContentResponseUsageMetadata,
    HarmCategory,
    MediaResolution,
    Modality,
    SchemaDict,
    SpeechConfigDict,
)
from haiway import ObservabilityLevel, as_list, ctx

from draive.gemini.api import GeminiAPI
from draive.gemini.config import GeminiGenerationConfig, GeminiSafetyConfig
from draive.gemini.lmm import context_element_as_content, result_part_as_content_or_call
from draive.gemini.types import GeminiException
from draive.gemini.utils import unwrap_missing
from draive.lmm import (
    LMM,
    LMMCompletion,
    LMMContext,
    LMMInstruction,
    LMMOutput,
    LMMOutputSelection,
    LMMStreamChunk,
    LMMStreamOutput,
    LMMToolRequest,
    LMMToolRequests,
    LMMTools,
)
from draive.lmm.helpers import lmm_output_decoder
from draive.lmm.types import LMMOutputDecoder
from draive.multimodal.content import MultimodalContent, MultimodalContentElement
from draive.utils import RateLimitError

__all__ = ("GeminiLMMGeneration",)


class GeminiLMMGeneration(GeminiAPI):
    def lmm(self) -> LMM:
        return LMM(completing=self.lmm_completion)

    @overload
    async def lmm_completion(
        self,
        *,
        instruction: LMMInstruction | None,
        context: LMMContext,
        tools: LMMTools | None,
        output: LMMOutputSelection,
        stream: Literal[False] = False,
        config: GeminiGenerationConfig | None = None,
        **extra: Any,
    ) -> LMMOutput: ...

    @overload
    async def lmm_completion(
        self,
        *,
        instruction: LMMInstruction | None,
        context: LMMContext,
        tools: LMMTools | None,
        output: LMMOutputSelection,
        stream: Literal[True],
        config: GeminiGenerationConfig | None = None,
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput]: ...

    async def lmm_completion(
        self,
        *,
        instruction: LMMInstruction | None,
        context: LMMContext,
        tools: LMMTools | None,
        output: LMMOutputSelection,
        stream: bool = False,
        config: GeminiGenerationConfig | None = None,
        **extra: Any,
    ) -> AsyncIterator[LMMStreamOutput] | LMMOutput:
        """
        Generate completion using Gemini model.

        Args:
            instruction: System instruction for the model
            context: Context elements for the conversation
            tools: Available tools for the model to use
            output: Output specification
            stream: Whether to stream the response
            config: Generation configuration
            **extra: Additional configuration parameters

        Returns:
            LMM output or async iterator of stream outputs
        """
        tools = tools or LMMTools.none
        generation_config: GeminiGenerationConfig = config or ctx.state(GeminiGenerationConfig)

        async with ctx.scope("gemini_lmm_completion", generation_config):
            request_config: GenerateContentConfigDict = _prepare_request_config(
                instruction=instruction,
                context=context,
                tools=tools,
                output=output,
                config=generation_config,
            )
            request_content: ContentListUnionDict = [
                context_element_as_content(element) for element in context
            ]
            output_decoder: LMMOutputDecoder = lmm_output_decoder(output)

            if stream:
                return await self._completion_stream(
                    model=generation_config.model,
                    content=request_content,
                    config=request_config,
                    output_decoder=output_decoder,
                )

            return await self._completion(
                model=generation_config.model,
                content=request_content,
                config=request_config,
                output_decoder=output_decoder,
            )

    async def _completion(  # noqa: C901, PLR0912
        self,
        model: str,
        content: ContentListUnionDict,
        config: GenerateContentConfigDict,
        output_decoder: Callable[[MultimodalContent], MultimodalContent],
    ) -> LMMOutput:
        try:
            completion: GenerateContentResponse = await self._client.aio.models.generate_content(
                model=model,
                config=config,
                contents=content,
            )

        except ResourceExhausted as exc:
            ctx.record(
                ObservabilityLevel.WARNING,
                event="lmm.rate_limit",
            )
            raise RateLimitError(
                retry_after=random.uniform(0.3, 3.0),  # nosec: B311
            ) from exc

        except Exception as exc:
            raise GeminiException(f"Failed to generate Gemini completion: {exc}") from exc

        _record_usage_metrics(
            completion.usage_metadata,
            model=model,
        )

        if not completion.candidates:
            ctx.record(
                ObservabilityLevel.ERROR,
                event="lmm.completion.error",
                attributes={"error": "Missing candidates"},
            )
            raise GeminiException("Invalid Gemini completion - missing candidates!", completion)

        completion_candidate: Candidate = completion.candidates[0]  # we always request only one

        if completion_candidate.safety_ratings:
            ctx.record(
                ObservabilityLevel.INFO,
                event="lmm.safety.results",
                attributes={
                    "results": [
                        f"{rating.category} |blocked: {rating.blocked}"
                        f" |probability:{rating.probability_score}"
                        f" |severity:{rating.severity_score}"
                        for rating in completion_candidate.safety_ratings
                        if rating.category
                    ],
                },
            )

        if completion_candidate.finish_reason == FinishReason.SAFETY:
            ctx.record(
                ObservabilityLevel.ERROR,
                event="lmm.safety.blocked",
                attributes={"reason": str(completion_candidate.finish_message)},
            )
            raise GeminiException("Gemini moderation blocked", completion)

        if not completion_candidate.content or not completion_candidate.content.parts:
            ctx.record(
                ObservabilityLevel.ERROR,
                event="lmm.completion.error",
                attributes={"error": "Missing content"},
            )
            raise GeminiException("Empty Gemini completion content!")

        result_content_elements: list[MultimodalContentElement] = []
        tool_requests: list[LMMToolRequest] = []
        for element in chain.from_iterable(
            result_part_as_content_or_call(part) for part in completion_candidate.content.parts
        ):
            if isinstance(element, LMMToolRequest):
                tool_requests.append(element)

            else:
                result_content_elements.append(element)

        completion_content: MultimodalContent
        if result_content_elements:
            completion_content = output_decoder(MultimodalContent.of(*result_content_elements))

        else:
            completion_content = MultimodalContent.empty

        if completion_candidate.finish_reason == FinishReason.MAX_TOKENS:
            ctx.record(
                ObservabilityLevel.WARNING,
                event="lmm.completion.warning",
                attributes={"finish_reason": str(completion_candidate.finish_message)},
            )

        elif completion_candidate.finish_reason != FinishReason.STOP:
            ctx.record(
                ObservabilityLevel.ERROR,
                event="lmm.completion.error",
                attributes={"finish_reason": str(completion_candidate.finish_message)},
            )
            raise GeminiException(f"Gemini completion error: {completion_candidate.finish_message}")

        if tool_requests:
            ctx.record(
                ObservabilityLevel.INFO,
                event="lmm.tool_requests",
                attributes={
                    "finish_reason": str(completion_candidate.finish_reason),
                    "lmm.tools": [call.tool for call in tool_requests],
                },
            )
            return LMMToolRequests.of(
                tool_requests,
                content=completion_content,
                meta={"finish_reason": str(completion_candidate.finish_reason)},
            )

        else:
            ctx.record(
                ObservabilityLevel.INFO,
                event="lmm.completion",
                attributes={"finish_reason": str(completion_candidate.finish_reason)},
            )
            return LMMCompletion.of(
                completion_content,
                meta={"finish_reason": str(completion_candidate.finish_reason)},
            )

    async def _completion_stream(
        self,
        model: str,
        content: ContentListUnionDict,
        config: GenerateContentConfigDict,
        output_decoder: LMMOutputDecoder,
    ) -> AsyncIterator[LMMStreamOutput]:
        completion_stream: AsyncIterator[GenerateContentResponse]
        try:
            completion_stream = await self._client.aio.models.generate_content_stream(
                model=model,
                config=config,
                contents=content,
            )

        except ResourceExhausted as exc:
            ctx.record(
                ObservabilityLevel.WARNING,
                event="lmm.rate_limit",
            )
            raise RateLimitError(
                retry_after=random.uniform(0.3, 3.0),  # nosec: B311
            ) from exc

        except Exception as exc:
            raise GeminiException(f"Failed to initialize Gemini streaming: {exc}") from exc

        return ctx.stream(
            _process_stream,
            stream=completion_stream,
            model=model,
            output_decoder=output_decoder,
        )


async def _process_stream(  # noqa PLR0912
    *,
    stream: AsyncIterator[GenerateContentResponse],
    model: str,
    output_decoder: LMMOutputDecoder,
) -> AsyncGenerator[LMMStreamChunk | LMMToolRequest]:
    async for chunk in stream:
        _record_usage_metrics(
            chunk.usage_metadata,
            model=model,
        )

        if not chunk.candidates:
            ctx.record(
                ObservabilityLevel.ERROR,
                event="lmm.completion.error",
                attributes={"error": "Missing candidates"},
            )
            raise GeminiException("Invalid Gemini chunk - missing candidates!", chunk)

        completion_candidate: Candidate = chunk.candidates[0]  # we always request only one

        if completion_candidate.safety_ratings:
            ctx.record(
                ObservabilityLevel.INFO,
                event="lmm.safety.results",
                attributes={
                    "results": [
                        f"{rating.category} |blocked: {rating.blocked}"
                        f" |probability:{rating.probability_score}"
                        f" |severity:{rating.severity_score}"
                        for rating in completion_candidate.safety_ratings
                        if rating.category
                    ],
                },
            )

        if completion_candidate.finish_reason == FinishReason.SAFETY:
            ctx.record(
                ObservabilityLevel.ERROR,
                event="lmm.safety.blocked",
                attributes={"finish_reason": str(completion_candidate.finish_message)},
            )
            raise GeminiException("Gemini moderation blocked", chunk)

        if not completion_candidate.content or not completion_candidate.content.parts:
            match completion_candidate.finish_reason:
                case None:
                    ctx.record(
                        ObservabilityLevel.WARNING,
                        event="lmm.completion.warning",
                        attributes={"warning": "Empty chunk content"},
                    )
                    continue  # skip empty content

                case FinishReason.STOP:
                    ctx.record(
                        ObservabilityLevel.INFO,
                        event="lmm.completion",
                        attributes={"finish_reason": str(completion_candidate.finish_message)},
                    )
                    yield LMMStreamChunk.of(
                        MultimodalContent.empty,
                        eod=True,
                        meta={"finish_reason": str(completion_candidate.finish_message)},
                    )
                    return  # end of stream

                case FinishReason.MAX_TOKENS:
                    ctx.record(
                        ObservabilityLevel.WARNING,
                        event="lmm.completion.warning",
                        attributes={"finish_reason": str(completion_candidate.finish_message)},
                    )
                    yield LMMStreamChunk.of(
                        MultimodalContent.empty,
                        eod=True,
                        meta={"finish_reason": str(completion_candidate.finish_message)},
                    )
                    return  # end of stream

                case other:
                    ctx.record(
                        ObservabilityLevel.ERROR,
                        event="lmm.completion.error",
                        attributes={"finish_reason": str(other)},
                    )
                    raise GeminiException(f"Gemini completion error: {other}")

        chunk_content: list[MultimodalContentElement] = []
        for part in completion_candidate.content.parts:
            for element in result_part_as_content_or_call(part):
                if isinstance(element, LMMToolRequest):
                    ctx.record(
                        ObservabilityLevel.INFO,
                        event="lmm.tool_request",
                        attributes={"lmm.tool": element.tool},
                    )
                    yield element

                else:
                    chunk_content.append(element)

        if completion_candidate.finish_reason == FinishReason.STOP:
            yield LMMStreamChunk.of(
                output_decoder(MultimodalContent.of(*chunk_content)),
                eod=True,
                meta={"finish_reason": str(completion_candidate.finish_reason)},
            )
            return  # end of stream

        elif completion_candidate.finish_reason == FinishReason.MAX_TOKENS:
            ctx.record(
                ObservabilityLevel.WARNING,
                event="lmm.completion.warning",
                attributes={"finish_reason": str(completion_candidate.finish_message)},
            )

            yield LMMStreamChunk.of(
                output_decoder(MultimodalContent.of(*chunk_content)),
                eod=True,
                meta={"finish_reason": str(completion_candidate.finish_reason)},
            )
            return  # end of stream

        elif completion_candidate.finish_reason is not None:
            ctx.record(
                ObservabilityLevel.ERROR,
                event="lmm.completion.error",
                attributes={"finish_reason": str(completion_candidate.finish_message)},
            )
            raise GeminiException(f"Gemini completion error: {completion_candidate.finish_message}")

        else:  # continue with chunk
            yield LMMStreamChunk.of(
                MultimodalContent.of(*chunk_content),
                eod=False,
            )


def _record_usage_metrics(
    usage: GenerateContentResponseUsageMetadata | None,
    *,
    model: str,
) -> None:
    if usage is None:
        return

    ctx.record(
        ObservabilityLevel.INFO,
        metric="lmm.input_tokens",
        value=usage.prompt_token_count or 0,
        unit="tokens",
        kind="counter",
        attributes={"lmm.model": model},
    )
    ctx.record(
        ObservabilityLevel.INFO,
        metric="lmm.input_tokens.cached",
        value=usage.cached_content_token_count or 0,
        unit="tokens",
        kind="counter",
        attributes={"lmm.model": model},
    )
    ctx.record(
        ObservabilityLevel.INFO,
        metric="lmm.output_tokens",
        value=usage.candidates_token_count or 0,
        unit="tokens",
        kind="counter",
        attributes={"lmm.model": model},
    )


def _prepare_request_config(  # noqa: C901, PLR0912, PLR0915
    *,
    instruction: LMMInstruction | None,
    context: LMMContext,
    tools: LMMTools,
    output: LMMOutputSelection,
    config: GeminiGenerationConfig,
) -> GenerateContentConfigDict:
    temperature: float | None = unwrap_missing(config.temperature)
    max_tokens: int | None = unwrap_missing(config.max_tokens)
    thinking_budget: int | None = unwrap_missing(config.thinking_budget)
    seed: int | None = unwrap_missing(config.seed)
    ctx.record(
        ObservabilityLevel.INFO,
        attributes={
            "lmm.provider": "gemini",
            "lmm.model": config.model,
            "lmm.temperature": temperature,
            "lmm.max_tokens": max_tokens,
            "lmm.thinking_budget": thinking_budget,
            "lmm.instruction": instruction,
            "lmm.context": [element.to_str() for element in context],
            "lmm.tools": [tool["name"] for tool in tools.specifications],
            "lmm.tool_selection": f"{tools.selection}",
            "lmm.output": f"{output}",
            "lmm.seed": seed,
        },
    )

    response_schema: SchemaDict | None
    response_modalities: list[Modality] | None
    response_mime_type: str | None
    match output:
        case "auto":
            # not specified at all - use defaults
            response_schema = None
            response_modalities = None
            response_mime_type = None

        case "text":
            response_schema = None
            response_modalities = [Modality.TEXT]
            response_mime_type = "text/plain"

        case "json":
            response_schema = None
            response_modalities = [Modality.TEXT]
            response_mime_type = "application/json"

        case "image":
            response_schema = None
            response_modalities = [
                Modality.TEXT,
                Modality.IMAGE,
            ]  # google api does not allow to specify only image
            response_mime_type = None  # define mime type?

        case "audio":
            response_schema = None
            response_modalities = [Modality.AUDIO]
            response_mime_type = None

        case "video":
            raise NotImplementedError("video output is not supported by Gemini")

        case ["text", "image"] | ["image", "text"]:  # refine multimodal matching?
            response_schema = None
            response_modalities = [
                Modality.TEXT,
                Modality.IMAGE,
            ]
            response_mime_type = None

        case [*_]:
            raise NotImplementedError("multimodal output is not supported by Gemini")

        case model:
            response_schema = cast(SchemaDict, model.__PARAMETERS_SPECIFICATION__)
            response_modalities = [Modality.TEXT]
            response_mime_type = "application/json"

    functions: list[FunctionDeclarationDict] | None
    function_calling_mode: FunctionCallingConfigMode | None
    allowed_function_names: list[str] | None = None
    if tools.specifications:
        functions = [
            FunctionDeclarationDict(
                name=tool["name"],
                description=tool["description"],
                parameters=cast(SchemaDict, tool["parameters"]),
            )
            for tool in tools.specifications or []
        ]
        match tools.selection:
            case "auto":
                function_calling_mode = FunctionCallingConfigMode.AUTO

            case "required":
                function_calling_mode = FunctionCallingConfigMode.ANY

            case "none":
                functions = None  # no need to pass functions if none can be used
                function_calling_mode = FunctionCallingConfigMode.NONE

            case specific_tool:  # handle specific tool selection
                function_calling_mode = FunctionCallingConfigMode.ANY
                allowed_function_names = [specific_tool["name"]]

    else:  # no functions
        functions = None
        function_calling_mode = FunctionCallingConfigMode.NONE

    media_resolution: MediaResolution | None
    match config.media_resolution:
        case "low":
            media_resolution = MediaResolution.MEDIA_RESOLUTION_LOW

        case "medium":
            media_resolution = MediaResolution.MEDIA_RESOLUTION_MEDIUM

        case "high":
            media_resolution = MediaResolution.MEDIA_RESOLUTION_HIGH

        case _:
            media_resolution = None

    speech_config: SpeechConfigDict | None
    if config.speech_voice_name:
        speech_config = {
            "voice_config": {
                "prebuilt_voice_config": {
                    "voice_name": cast(str, config.speech_voice_name),
                },
            }
        }

    else:
        speech_config = None

    safety_config: GeminiSafetyConfig = unwrap_missing(
        config.safety,
        default=GeminiSafetyConfig(),
    )

    return {
        "temperature": temperature,
        "top_p": unwrap_missing(config.top_p),
        "top_k": unwrap_missing(config.top_k),
        "max_output_tokens": max_tokens,
        "candidate_count": 1,
        "seed": seed,
        "stop_sequences": unwrap_missing(
            config.stop_sequences,
            transform=as_list,
        ),
        "safety_settings": [
            {
                "category": HarmCategory.HARM_CATEGORY_HATE_SPEECH,
                "threshold": safety_config.harm_category_hate_speech_threshold,
            },
            {
                "category": HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT,
                "threshold": safety_config.harm_category_dangerous_content_threshold,
            },
            {
                "category": HarmCategory.HARM_CATEGORY_HARASSMENT,
                "threshold": safety_config.harm_category_harassment_threshold,
            },
            {
                "category": HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT,
                "threshold": safety_config.harm_category_sexually_explicit_threshold,
            },
            {
                "category": HarmCategory.HARM_CATEGORY_CIVIC_INTEGRITY,
                "threshold": safety_config.harm_category_civic_integrity_threshold,
            },
        ],
        "system_instruction": instruction,
        "tools": [{"function_declarations": functions}] if functions else None,
        "tool_config": {
            "function_calling_config": {
                "mode": function_calling_mode,
                **(
                    {"allowed_function_names": allowed_function_names}
                    if allowed_function_names
                    else {}
                ),
            }
        }
        if function_calling_mode
        else None,
        # prevent google from automatically resolving function calls
        "automatic_function_calling": {
            "disable": True,
            "maximum_remote_calls": None,
        }
        if function_calling_mode
        else None,
        "media_resolution": media_resolution,
        "response_modalities": response_modalities,
        "response_mime_type": response_mime_type,
        "response_schema": response_schema,
        "speech_config": speech_config,
        "thinking_config": {
            "include_thoughts": True,
            "thinking_budget": thinking_budget,
        }
        if thinking_budget is not None
        else None,
    }
